import phanterpwa.frontend.gatehandler as gatehandler
import phanterpwa.frontend.helpers as helpers

from org.transcrypt.stubs.browser import __pragma__

__pragma__('alias', "jQuery", "$")
__pragma__('skip')
jQuery = sessionStorage = JSON = M = js_undefined = window = setTimeout = \
    __new__ = FormData = console = localStorage = document = this = 0
__pragma__('noskip')

DIV = helpers.XmlConstructor.tagger("div")
I = helpers.XmlConstructor.tagger("i")
H1 = helpers.XmlConstructor.tagger("h1")
H2 = helpers.XmlConstructor.tagger("h2")
H3 = helpers.XmlConstructor.tagger("h3")
STRONG = helpers.XmlConstructor.tagger("strong")
SPAN = helpers.XmlConstructor.tagger("span")
A = helpers.XmlConstructor.tagger("a")
UL = helpers.XmlConstructor.tagger("ul")
LI = helpers.XmlConstructor.tagger("li")
LABEL = helpers.XmlConstructor.tagger("label")
XSECTION = helpers.XSECTION
I18N = helpers.I18N
CONCATENATE = helpers.CONCATENATE


__pragma__('kwargs')


class Index(gatehandler.Handler):
    def __init__(self, request):
        gatehandler.Handler.__init__(self, request)

    def initialize(self):
        html = CONCATENATE(
            DIV(
                DIV(
                    DIV(
                        DIV("TEST REQUERIMENTS", _class="phanterpwa-breadcrumb"),
                        _class="phanterpwa-breadcrumb-wrapper"
                    ),
                    _class="p-container"),
                _class='title_page_container card'
            ),
            DIV(
                DIV(
                    DIV(
                        DIV(
                            _style="width:100%; text-align: center; padding-top: 100px; padding-bottom: 100px;"
                        ),
                        _id="requeriments_container"),
                    _class="card p-row e-padding_10"
                ),
                _class="phanterpwa-container p-container"
            )
        )
        html.html_to("#main-container")

    def _binds(self):
        jQuery("#phanterpwa-component-left_bar-menu_button-change_projects_folder").off(
            "click.change_projects_folder"
        ).on(
            "click.change_projects_folder",
            self.change_projects_folder,
        )

    def reload(self):
        self._binds()

    def _bind_open_api(self, app_name, url):

        jQuery("#phanterpwa-component-left_bar-menu_button-view_api").off(
            "click.open_href_play"
        ).on(
            "click.open_href_play",
            lambda: window.open(url, "_blank")
        )



__pragma__('nokwargs')



import os
import asyncio
import sys
import re
import time
import importlib
import subprocess
import threading
import psutil
import phanterpwa
import configparser
from tornado import (
    web,
    ioloop,
    httpserver,
    autoreload,
    netutil,
    process
)
from phanterpwa.tools import config
from phanterpwa import compiler
from phanterpwa import configer

CURRENT_DIR = os.path.join(os.path.dirname(__file__))
PHANTERPWA_VERSION = phanterpwa.__version__

version_regex = re.compile(r"^[0-9]{1,}\.[0-9]{1,}\.[0-9]{1,}$")


class PhanterPWATornado(object):
    """docstring for PhanterPWATornado"""

    def __init__(self, projectPath):
        super(PhanterPWATornado, self).__init__()
        self.projectPath = projectPath
        self.projectConfig = configer.ProjectConfig(os.path.join(self.projectPath, "config.json"))
        self.apps_ports = []
        if self.projectConfig.get("FRONTEND") and \
            os.path.exists(os.path.join(self.projectConfig['PROJECT']['path'], "frontapps")):
            for x in self.projectConfig['FRONTEND']:
                current_port = self.projectConfig['FRONTEND'][x]['port']
                if current_port not in self.apps_ports:
                    self.apps_ports.append(current_port)
                else:
                    raise ValueError("The '{0}' frontend app port ({1}) is not valid".format(x, current_port))

        if self.projectConfig.get("BACKEND") and \
            os.path.exists(os.path.join(self.projectConfig['PROJECT']['path'], "backapps")):
            for x in self.projectConfig['BACKEND']:
                current_port = self.projectConfig['BACKEND'][x]['port']
                if current_port not in self.apps_ports:
                    self.apps_ports.append(current_port)
                else:
                    raise ValueError("The '{0}' backend app port ({1}) is not valid".format(x, current_port))

        self.debug = self.projectConfig["PROJECT"]["debug"]
        self.version = self.projectConfig["PROJECT"]["version"]
        self.oncli = False

    @property
    def oncli(self):
        return self._oncli

    @oncli.setter
    def oncli(self, oncli):
        if isinstance(oncli, bool):
            self._oncli = oncli
        else:
            raise "The oncli must be boolean. Given: {0}".format(type(oncli))

    def compile(self):
        c = compiler.Compiler(self.projectPath)
        c.compile()

    def run(self):
        sys.path.insert(0, self.projectPath)
        os.chdir(self.projectPath)
        print(self.projectPath)

        if self.apps_ports:
            self.cpu_number = int(self.projectConfig["PROJECT"].get('cpu_number', 1))
            if os.name == "nt":
                self.cpu_number = 1
            needs_fork = self.cpu_number != 1 and self.cpu_number > -1
            if needs_fork:

                socs = {}
                for x in self.projectConfig['BACKEND']:
                    current_port = self.projectConfig['BACKEND'][x]['port']
                    socket = netutil.bind_sockets(current_port)
                    socs[f'back_{x}'] = socket
                for x in self.projectConfig['FRONTEND']:
                    current_port = self.projectConfig['FRONTEND'][x]['port']
                    socket = netutil.bind_sockets(current_port)
                    socs[f'front_{x}'] = socket
                time.sleep(2)
                process.fork_processes(self.cpu_number)
                time.sleep(2 * (os.getpid() % self.cpu_number))
                print(f"After fork - PID: {os.getpid()}")

                # ADIÇÃO: Função com delay para inicialização de apps
                async def initialize_apps_with_delay():
                    delay_between_apps = 1.5  # 200ms entre cada app

                    # Inicializar BACKEND apps com delay
                    backend_apps = list(self.projectConfig['BACKEND'].keys())
                    for i, x in enumerate(backend_apps):
                        await asyncio.sleep(i * delay_between_apps)
                        self._initialize_backend_app(x, socs[f'back_{x}'])
                        print(f"Backend app {x} initialized with delay")

                    # Inicializar FRONTEND apps com delay
                    frontend_apps = list(self.projectConfig['FRONTEND'].keys())
                    for i, x in enumerate(frontend_apps):
                        await asyncio.sleep(i * delay_between_apps)
                        self._initialize_frontend_app(x, socs[f'front_{x}'])
                        print(f"Frontend app {x} initialized with delay")

                # Executar inicialização com delays
                ioloop.IOLoop.current().add_callback(initialize_apps_with_delay)
            else:
                for x in self.projectConfig['BACKEND']:
                    current_port = self.projectConfig['BACKEND'][x]['port']
                    handlers_app = importlib.import_module("backapps.{0}.handlers".format(x))
                    if isinstance(handlers_app.HANDLER, (list, tuple)):
                        app = web.Application(
                            handlers_app.HANDLER,
                            **handlers_app.SETTINGS
                        )
                    else:
                        app = handlers_app.HANDLER

                    app_http_server = httpserver.HTTPServer(app)
                    app_http_server.listen(current_port)
                    print(id(app_http_server))

                for x in self.projectConfig['FRONTEND']:
                    current_port = self.projectConfig['FRONTEND'][x]['port']
                    handlers_app = importlib.import_module("frontapps.{0}.handlers".format(x))
                    if isinstance(handlers_app.HANDLER, (list, tuple)):
                        app = web.Application(
                            handlers_app.HANDLER,
                            **handlers_app.SETTINGS
                        )
                    else:
                        app = handlers_app.HANDLER

                    app_http_server = httpserver.HTTPServer(app)
                    app_http_server.listen(current_port)
                    print(id(app_http_server))

        autoreload.watch(os.path.join(self.projectPath, "config.json"))
        ioloop.IOLoop.current().start()
        print("start stopped")
        ioloop.IOLoop.current().add_callback(lambda: ioloop.IOLoop.current().close(True))
        print("close? Don't know")

    def _initialize_backend_app(self, app_name, socket):
        """Inicializa uma aplicação backend com tratamento de erro"""
        try:
            time.sleep(0.5)
            handlers_app = importlib.import_module("backapps.{0}.handlers".format(app_name))
            time.sleep(0.3)
            if isinstance(handlers_app.HANDLER, (list, tuple)):
                app = web.Application(
                    handlers_app.HANDLER,
                    **handlers_app.SETTINGS
                )
            else:
                app = handlers_app.HANDLER
            time.sleep(0.3)
            app_http_server = httpserver.HTTPServer(app)
            app_http_server.add_sockets(socket)
            time.sleep(0.2)
            return True
        except Exception as e:
            print(f"Error initializing backend app {app_name}: {e}")
            return False

    def _initialize_frontend_app(self, app_name, socket):
        """Inicializa uma aplicação frontend com tratamento de erro"""
        try:
            time.sleep(0.5)
            handlers_app = importlib.import_module("frontapps.{0}.handlers".format(app_name))
            time.sleep(0.3)
            if isinstance(handlers_app.HANDLER, (list, tuple)):
                app = web.Application(
                    handlers_app.HANDLER,
                    **handlers_app.SETTINGS
                )
            else:
                app = handlers_app.HANDLER
            time.sleep(0.5)
            app_http_server = httpserver.HTTPServer(app)
            app_http_server.add_sockets(socket)
            time.sleep(0.2)
            return True
        except Exception as e:
            print(f"Error initializing frontend app {app_name}: {e}")
            return False

    def stop(self):
        ioloop.IOLoop.current().add_callback(
            lambda: (
                ioloop.IOLoop.current().stop()
            )
        )
        print("comando stop")


class ProjectRunner():
    def __init__(self):
        self.path_phanterpwa = CURRENT_DIR
        self.env_python = os.path.normpath(sys.executable)
        self._projects_dict = {}
        self._ini_project = configparser.ConfigParser()

    @property
    def projects(self):
        self._projects()
        return self._projects_dict

    def run(self, project_path, compile=False, thread=False):
        temp_dir = os.path.join(project_path, "temp")
        ppwa_version_file = os.path.join(project_path, "temp", 'ppwa_version')
        ppwa_version = None
        project_version = "0.0.1"

        if os.path.isfile(ppwa_version_file):
            with open(ppwa_version_file, 'r', encoding="utf-8") as f:
                ppwa_version = f.read()
        elif os.path.isdir(temp_dir):
            ppwa_version = PHANTERPWA_VERSION
            with open(ppwa_version_file, 'w', encoding="utf-8") as f:
                f.write(ppwa_version)            
        if os.path.isfile(os.path.join(project_path, 'project.ini')):
            self._ini_project.read(
                os.path.join(project_path, 'project.ini'), encoding='utf-8')
            project_version = self._ini_project["PROJECT"].get("version", "0.0.1")

        if os.path.isfile(os.path.join(project_path, "config.json")):
            cfg_ini = config(os.path.join(project_path, "config.json"))
        else:
            cfg_ini = configer.ProjectConfig(project_path)
        project_version_cfg = cfg_ini["PROJECT"]["version"]
        reset_compilation = False
        cfg = configer.ProjectConfig(project_path)
        if str(project_version).strip() != str(project_version_cfg).strip():
            reset_compilation = True
      

        if compile:
            c = compiler.Compiler(project_path, reset_compilation=reset_compilation)
            c.compile()
        else:
            f_comp = False
            if str(project_version).strip() != str(project_version_cfg).strip():
                f_comp = True
                print("Project version change:", str(project_version_cfg).strip(), "-->", str(project_version).strip())
            if str(ppwa_version).strip() != str(PHANTERPWA_VERSION).strip():
                f_comp = True
                print("PhanterPWA version change:",
                    str(ppwa_version).strip(), "-->", str(PHANTERPWA_VERSION).strip())
            if f_comp:
                c = compiler.Compiler(project_path, reset_compilation=reset_compilation)
                c.compile()
   
        with open(ppwa_version_file, 'w', encoding="utf-8") as f:
            f.write(PHANTERPWA_VERSION)
        target = os.path.normpath(os.path.join(self.path_phanterpwa, "server.py {0}".format(project_path)))
        self.stop(project_path)
        command = " ".join([self.env_python, "-X utf8", target])
        print()
        print("=" * 79)
        print("Starting server..........")
        print("CPUs:", cfg["PROJECT"].get('cpu_number') or os.cpu_count())
        if os.name == "nt":
            print("CPUs:", 1)
        if cfg['BACKEND']:
            print('BACKEND')
            for a in cfg['BACKEND']:
                print(" ", a)
                print("    HOST:", cfg['BACKEND'][a]['host'])
                print("    PORT:", cfg['BACKEND'][a]['port'])

        print()
        if cfg['FRONTEND']:
            print('FRONTEND')
            for a in cfg['FRONTEND']:
                print(" ", a)
                print("    HOST:", cfg['FRONTEND'][a]['host'])
                print("    PORT:", cfg['FRONTEND'][a]['port'])
        print()
        with open(os.path.join(self.path_phanterpwa, "samples", "art"), "r") as f:
            print(f.read())
        print()
        print("=" * 79)
        if thread:
            t = threading.Thread(target=lambda: subprocess.call(command, cwd=project_path, shell=True))
            t.start()
            return t
        else:
            try:
                subprocess.call(command, cwd=project_path, shell=True)
            except KeyboardInterrupt:
                self.stop(project_path)

    def stop_all(self):
        all_p = self.projects
        for p in all_p:
            for x in all_p[p]:
                x.terminate()

    def stop(self, project_path):
        has_running = False
        for x in self._project(project_path):
            has_running = True
            x.terminate()
        self._projects()
        if has_running:
            print("=" * 79)
            print("Stoping server....")
            cfg = configer.ProjectConfig(project_path)
            if cfg['BACKEND']:
                print('BACKEND')
                for a in cfg['BACKEND']:
                    print(" ", a)
                    print("    HOST:", cfg['BACKEND'][a]['host'])
                    print("    PORT:", cfg['BACKEND'][a]['port'])
            print()
            if cfg['FRONTEND']:
                print('FRONTEND')
                for a in cfg['FRONTEND']:
                    print(" ", a)
                    print("    HOST:", cfg['FRONTEND'][a]['host'])
                    print("    PORT:", cfg['FRONTEND'][a]['port'])
            print()
            print("Goodbye!")

    def check(self, project_path):
        self._projects()
        if project_path in self._projects_dict:
            return True
        else:
            return False

    def _project(self, project_path):
        self._projects()
        if project_path in self._projects_dict:
            return self._projects_dict[project_path]
        else:
            return []

    def _projects(self):
        target = os.path.normpath(os.path.join(self.path_phanterpwa, "server.py"))
        self._projects_dict = {}
        for p in psutil.process_iter():
            cmd_line = None
            try:
                cmd_line = p.cmdline()
            except Exception:
                pass
            if cmd_line:
                if self.env_python == cmd_line[0]:
                    if len(cmd_line) > 1 and target in cmd_line:
                        if cmd_line[-1] in self._projects_dict:
                            self._projects_dict[cmd_line[-1]].append(p)
                        else:
                            self._projects_dict[cmd_line[-1]] = [p]


if __name__ == "__main__":
    import os
    projectPath = os.getcwd()
    AppRunv = PhanterPWATornado(projectPath)
    try:
        AppRunv.run()
    except KeyboardInterrupt:
        AppRunv.stop()




from ..helpers import (
    DIV,
    LABEL,
    INPUT,
    I,
    # SPAN,
    TEXTAREA,
    SELECT,
    OPTION,
    # CONCATENATE,
    # UL,
    # LI,
    # A
)
from ..xmlconstructor import XmlConstructor


class BASEINPUT(XmlConstructor):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        input_type: "type of input"="text",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):

        attributes["_type"] = input_type
        attributes["_name"] = name
        if "_id" not in attributes:
            attributes["_id"] = "phanterpwa-input-%s" % name

        if "_title" not in attributes:
            attributes["_title"] = label
        attrs_container = {
            "_id": "phanterpwa-container-input-text-%s" % name,
            "_class": "phanterpwa-container-input-text phanterpwa-flag-%s" % style,
            "_phanterpwa-css_style": style,
            "_phanterpwa-input_name": name,
            "_phanterpwa-input_id": attributes["_id"],
        }
        if attributes_container:
            for x in attributes_container:
                if x == "_class":
                    attrs_container["_class"] = "phanterpwa-container-input-text phanterpwa-flag-%s %s" %\
                        (style, attributes_container[x])
                else:
                    attrs_container[x] = attributes_container[x]

        label_attr = {"_for": attributes["_id"]}

        if "_phanterpwa_languages" in attributes:
            label_attr['_phanterpwa_languages'] = attributes["_phanterpwa_languages"]
            attributes["_phanterpwa_languages"] = None

        if "_class" in attributes and attributes["_class"]:
            new_class = "form-control {0}".format(attributes["_class"].strip())
            if style == "bootstrap":
                new_class = " ".join([attributes["_class"].strip(), "form-control"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "form-control"
        self.input = INPUT(
            **attributes
        )
        if style == "materializecss":
            new_content = [
                DIV(I(_class="fas fa-check"),
                    _id="phanterpwa-input-check-%s" % name,
                    _class="phanterpwa-input-check"),
                DIV(
                    self.input,
                    LABEL(
                        label,
                        **label_attr),
                    _class="input-field"),
                DIV(error if error else "",
                    _title=error,
                    _id="phanterpwa-materialize-input-error-%s" % name,
                    _class="phanterpwa-materialize-input-error"),
            ]
        elif style == "bootstrap":
            new_content = [
                DIV(I(_class="fas fa-check"),
                    _id="phanterpwa-input-check-%s" % name,
                    _class="phanterpwa-input-check"),
                DIV(
                    LABEL(
                        label,
                        **label_attr),
                    self.input,
                    _class="form-group"),
                DIV(error if error else "",
                    _title=error,
                    _id="phanterpwa-materialize-input-error-%s" % name,
                    _class="phanterpwa-materialize-input-error"),
            ]
        else:
            raise ValueError("The style must be materializecss or bootstrap. given: %s" % style)
        XmlConstructor.__init__(self, 'div', False, *new_content, **attrs_container)


class InputTextarea(XmlConstructor):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        input_type: "type of input"="text",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):

        attributes["_type"] = input_type
        attributes["_name"] = name
        if "_id" not in attributes:
            attributes["_id"] = "phanterpwa-input-%s" % name

        if "_title" not in attributes:
            attributes["_title"] = label
        attrs_container = {
            "_id": "phanterpwa-container-input-text-%s" % name,
            "_class": "phanterpwa-container-input-text phanterpwa-flag-%s" % style,
            "_phanterpwa-css_style": style,
            "_phanterpwa-input_name": name,
            "_phanterpwa-input_id": attributes["_id"],
        }
        if attributes_container:
            for x in attributes_container:
                if x == "_class":
                    attrs_container["_class"] = "phanterpwa-container-input-text phanterpwa-flag-%s %s" %\
                        (style, attributes_container[x])
                else:
                    attrs_container[x] = attributes_container[x]

        label_attr = {"_for": attributes["_id"]}

        if "_phanterpwa_languages" in attributes:
            label_attr['_phanterpwa_languages'] = attributes["_phanterpwa_languages"]
            attributes["_phanterpwa_languages"] = None
        if style == "materializecss":
            if "_class" in attributes and attributes["_class"]:
                new_class = " ".join([attributes["_class"].strip(), "materialize-textarea"])
                attributes["_class"] = new_class
            else:
                attributes["_class"] = "materialize-textarea"

        elif style == "bootstrap":
            attributes["_rows"]
            if "_class" in attributes and attributes["_class"]:
                new_class = " ".join([attributes["_class"].strip(), "form-control"])
                attributes["_class"] = new_class
            else:
                attributes["_class"] = "form-control"
            if "_rows" not in attributes and attributes["_class"]:
                attributes["_rows"] = "3"

        self.input = TEXTAREA(
            **attributes
        )
        if style == "materializecss":
            new_content = [
                DIV(I(_class="fas fa-check"),
                    _id="phanterpwa-input-check-%s" % name,
                    _class="phanterpwa-input-check"),
                DIV(
                    self.input,
                    LABEL(
                        label,
                        **label_attr),
                    _class="input-field"),
                DIV(error if error else "",
                    _title=error,
                    _id="phanterpwa-materialize-input-error-%s" % name,
                    _class="phanterpwa-materialize-input-error"),
            ]
        elif style == "bootstrap":
            new_content = [
                DIV(I(_class="fas fa-check"),
                    _id="phanterpwa-input-check-%s" % name,
                    _class="phanterpwa-input-check"),
                DIV(
                    LABEL(
                        label,
                        **label_attr),
                    self.input,
                    _class="form-group"),
                DIV(error if error else "",
                    _title=error,
                    _id="phanterpwa-materialize-input-error-%s" % name,
                    _class="phanterpwa-materialize-input-error"),
            ]
        else:
            raise ValueError("The style must be materializecss or bootstrap. given: %s" % style)
        XmlConstructor.__init__(self, 'div', False, *new_content, **attrs_container)


class InputText(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "text"
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)


class InputSelect(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        placeholder="",
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "text"
        attributes['_phanterpwa-input-select'] = name
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)


class InputPassword(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "password"
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)


class InputHidden(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "hidden"
        if "_style" not in attributes_container:
            attributes_container["_style"] = "display: none;"
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)


class InputTextAndButton(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        icon: "Button Icon"=I(_class='fas fa-yin-yang'),
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "text"
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)
        old_content = self.content
        self.content = [
            DIV(
                old_content,
                DIV(
                    icon,
                    _id='phanterpwa_input_icon_button_{0}'.format(name),
                    _class='btn phanterpwa_input_icon_button'),
                _class="phanterpwa_input_icon_button_container"
            )
        ]


class InputSelectAndButton(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        icon: "Button Icon"=I(_class='fas fa-yin-yang'),
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "text"
        attributes['_phanterpwa-input-select'] = name
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)
        old_content = self.content
        self.content = [
            DIV(
                old_content,
                DIV(
                    icon,
                    _id='phanterpwa_input_icon_button_{0}'.format(name),
                    _class='btn phanterpwa_input_icon_button'),
                _class="phanterpwa_input_icon_button_container"
            )
        ]


class InputTextPlus(BASEINPUT):
    def __init__(
        self,
        name: "Name of input element",
        label: "Label of the element",
        error: "Message error on validation"=None,
        style: "Css framework style (materializecss or bootstrap)"="materializecss",
        attributes_container: "Attributes of container. The key must stats with underline (_)." = {},
        **attributes: "Attributes of input element. The key must stats with underline (_)."):
        input_type = "text"
        BASEINPUT.__init__(self, name, label, error, style, input_type, attributes_container, **attributes)
        old_content = self.content
        self.content = [
            DIV(
                old_content,
                DIV(
                    I(_class='fas fa-plus'),
                    _id='phanterpwa_input_plus_button_{0}'.format(name),
                    _class='btn phanterpwa_input_plus_button'),
                _class="phanterpwa_input_text_plus_container"
            )
        ]




import json
import types
import re
import time
import datetime
from phanterpwa.helpers import (
    FORM,
    DIV,
    I,
    INPUT,
    LABEL,
    BUTTON
)

from pydal.objects import (Set, Table, Field)
from pydal.validators import (
    IS_NOT_EMPTY,
    IS_DATE,
    IS_DATETIME,
    IS_TIME,
    IS_IN_SET,
    IS_EMAIL,
    IS_EMPTY_OR,
    IS_IN_DB
)

from phanterpwa.backend.pydal.extra_validations import IS_ACTIVATION_CODE

ref_values = re.compile(r"\$[0-9]{13}\:.*")


def datetime_formater(value):
    value = value.replace("%d", "dd").replace("%m", "MM").replace("%Y", "yyyy").replace(
        "%H", "HH").replace("%M", "mm").replace("%S", "ss")
    return value


def datetime_converter(iso_value, pattern="date", out_format="%Y-%m-%d %H:%M:%S", omit_errors=True):
    """
    @iso_value: The iso-value must be in the following format:
        "%Y-%m-%d %H:%M:%S" to date pattern

    @patter: There are three valid values: date, datetime and time
    @out_format: out strftime format
    @omit_errors: If it's True, on error returns None.
    """
    if pattern == "date":
        in_format = "%Y-%m-%d"
    elif pattern == "time":
        in_format = "%H:%M:%S"
    elif pattern == "datetime":
        in_format = "%Y-%m-%d %H:%M:%S"
    else:
        if omit_errors:
            print("the pattern is invalid. Given: {0}".format(pattern))
            return None
        else:
            raise ValueError("The pattern must be data, datetime or time")

    if omit_errors:
        try:
            strptime = time.strptime(str(iso_value), in_format)
            result = time.strftime(out_format, strptime)
        except Exception as e:
            print(e)
            return None
        else:
            return result
    else:
        strptime = time.strptime(str(iso_value), in_format)
        result = time.strftime(out_format, strptime)

    return result


class DataForm():
    _form_list = []

    def __init__(self, identifier, **parameters):
        DataForm._add_in_form_list(identifier)
        self.identifier = identifier
        self.table = parameters.get("table", None)
        self._id = parameters.get("id", None)
        self.uri = parameters.get("uri", None)

    def as_dict(self):
        FormFromTableDAL(self.table, self._id)

    @classmethod
    def _add_in_form_list(cls, identifier):
        cls._form_list.append(identifier)

    @property
    def form_list(self):
        return DataForm._form_list


class WidgetFromFieldDALFromTableDAL():
    def __init__(self, table, field, record_id=None, data_view=False, simple_widget=False, custom_phanterpwa=None):
        self.table = table
        self._db = self.table._db
        self._data_view = data_view
        self._field = field
        self._record = None
        self.record_id = record_id
        self.section = None
        self.position = None
        self.group = None
        self._custom_label = None
        self.out_of_form = False
        self.simple_widget = simple_widget
        self.phanterpwa = dict()
        if custom_phanterpwa:
            self.phanterpwa = custom_phanterpwa

        if hasattr(self.table[self._field], "phanterpwa"):
            for x in self.table[self._field].phanterpwa:
                if x not in self.phanterpwa:
                    self.phanterpwa[x] = self.table[self._field].phanterpwa[x]

        if "data_view" in self.phanterpwa:
            self._data_view = self.phanterpwa["data_view"]
        if "label" in self.phanterpwa:
            self._custom_label = self.phanterpwa["label"]
        if "position" in self.phanterpwa:
            self.position = self.phanterpwa["position"]
        if "section" in self.phanterpwa:
            self.section = self.phanterpwa["section"]
        elif "group" in self.phanterpwa:
            self.group = self.phanterpwa["group"]
        if "out_of_form" in self.phanterpwa and\
                self.phanterpwa["out_of_form"] is True:
            self.out_of_form = True
        if "simple_widget" in self.phanterpwa and\
                self.phanterpwa["simple_widget"] is True:
            self.simple_widget = True


        if "_class" not in self.phanterpwa:
            self.phanterpwa["_class"] = "p-col w1p100"

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if value is None:
            self._position = None
        elif isinstance(value, int):
            self._position = value
        elif str(value).isdigit():
            self._position = int(value)
        else:
            raise SyntaxError(
                "The position value must be intenger. Give: {0}".format(type(value)))

    @property
    def record(self):
        return self._record

    @property
    def record_id(self):
        return self._record_id

    @record_id.setter
    def record_id(self, value):
        self._record_id = None
        if value:
            r = self._db(self._db[self.table._tablename]._id == value).select().first()
            if r:
                self._record = r
                self._record_id = r.id
            else:
                raise ValueError("The record_id not find in table")

    @property
    def db(self):
        return self._db

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        if isinstance(value, Table):
            self._table = value
        else:
            raise ValueError("The table must be a pydal Table instance.")

    def __ppwa_check_validators(self, validator):

        if isinstance(validator, IS_NOT_EMPTY):
            if "can_empty" not in self.phanterpwa:
                self.phanterpwa["can_empty"] = False
            return ["IS_NOT_EMPTY"]
        elif isinstance(validator, IS_DATETIME):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_DATETIME:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_DATE):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_DATE:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_TIME):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_TIME:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_ACTIVATION_CODE):
            return ["IS_ACTIVATION_CODE"]
        elif isinstance(validator, IS_EMAIL):
            return ["IS_EMAIL"]
        elif isinstance(validator, IS_IN_DB):
            if "data_set" not in self.phanterpwa:
                validator.build_set()
                table_name = validator.ktable
                if callable(self._db[table_name]._format):
                    theset = [self._db[table_name]._format(x) for x in self._db(self._db[table_name].id > 0).select()]
                elif self._db[table_name]._format and isinstance(self._db[table_name]._format, str):
                    theset = [self._db[table_name]._format % x.as_dict() for x in self._db(self._db[table_name].id > 0).select()]
                else:
                    theset = [x.id for x in self._db(self._db[table_name].id > 0).select()]
                return ["IS_IN_SET:{0}".format(json.dumps("{0}".format(theset)))]
            else:
                if callable(self.phanterpwa['data_set']):
                    theset = self.phanterpwa['data_set'](self.record)
                else:
                    theset = self.phanterpwa['data_set']
                self.phanterpwa['data_set'] = theset
                return ["IS_IN_SET:{0}".format(json.dumps("{0}".format(theset)))]
        elif isinstance(validator, IS_IN_SET):
            if "data_set" not in self.phanterpwa and "type" not in self.phanterpwa:
                self.phanterpwa["type"] = "select"
                self.phanterpwa["data_set"] = list(validator.theset)
            return ["IS_IN_SET:{0}".format(json.dumps("{0}".format(validator.theset)))]
        elif isinstance(validator, IS_EMPTY_OR):
            if "can_empty" not in self.phanterpwa and \
                    "validators" not in self.phanterpwa:
                self.phanterpwa["can_empty"] = True
            elif "validators" in self.phanterpwa and \
                    "IS_NOT_EMPTY" in self.phanterpwa["validators"]:
                self.phanterpwa["can_empty"] = False
            other = self.__ppwa_check_validators(validator.other)
            if other:
                if isinstance(other, (list, tuple)):
                    return ["IS_EMPTY_OR", *other]
                else:
                    return ["IS_EMPTY_OR", other]
            else:
                return []
        elif isinstance(validator, type(lambda: 0)):
            return []
        elif isinstance(validator, list):
            return [self.__ppwa_check_validators(x) for x in validator]

    def as_dict(self):
        json_field = dict()
        if not self.out_of_form:
            FieldInst = self.table[self._field]
            t = FieldInst.type
            def u_list(v_list):
                result = []
                if isinstance(v_list, list):
                    t = list(v_list)
                    for item in v_list:
                        if (isinstance(item, (list, tuple))):
                            result.extend(item)
                        elif item:
                            result.append(item)
                return result

            auto_validators = u_list(self.__ppwa_check_validators(FieldInst.requires))
            if "validators" not in self.phanterpwa:
                json_field["validators"] = auto_validators
            else:
                json_field["validators"] = self.phanterpwa["validators"]
            default = FieldInst.default
            if "can_empty" in self.phanterpwa:
                json_field["can_empty"] = self.phanterpwa["can_empty"]

            if t.startswith("reference"):

                if self._record_id:
                    if self._record[FieldInst]:
                        default = self._record[FieldInst].id
                    else:
                        default = None
                json_field['editable'] = self.phanterpwa.get("editable", False)
                ref_table = t.split(" ")[1]
                if "data_set" in self.phanterpwa:
                    if callable(self.phanterpwa['data_set']):
                        data_ref_table_formated = self.phanterpwa['data_set'](self.record)
                    else:
                        data_ref_table_formated = self.phanterpwa["data_set"]
                    if json_field['editable']:
                        if 'reference_field' not in self.phanterpwa:
                            raise SyntaxError("".join(["You placed in '", self._field,
                                "' the reference field as editable, it is necessary to",
                                    " define the field of the referenced table. e.g. ",
                                        "phanterpwa = {'editable': True, 'refence_field': 'field_of_referenced_table'}"]))
                        else:
                            json_field['reference_field'] = self.phanterpwa["reference_field"]
                    else:
                        json_field['reference_field'] = 'id'
                    json_field['reference_table'] = ref_table
                else:
                    if not self.db:
                        raise SyntaxError("The db must be DAL database instance")
                    ref_fields = []
                    for f in self.db[ref_table].fields:
                        ref_fields.append(f)
                    data_ref_table = []
                    data_ref_table_formated = []
                    q_ref_table = self.db(self.db[ref_table]).select()
                    fmt = None
                    if self.db[ref_table]._format:
                        fmt = self.db[ref_table]._format
                    for q in q_ref_table:
                        f_format = {}
                        row = []
                        for z in ref_fields:
                            if (self.db[ref_table][z].type == "datetime") or\
                                    (self.db[ref_table][z].type == "date") or\
                                    (self.db[ref_table][z].type == "time"):
                                if q[z]:
                                    row.append(q[z].isoformat())
                                else:
                                    row.append(q[z])
                            else:
                                row.append(q[z])
                            f_format[z] = q[z]
                        data_ref_table.append(row)
                        if fmt:
                            if hasattr(fmt, '__call__'):
                                data_ref_table_formated.append(
                                    [q.id, fmt(q) % f_format])
                            else:
                                data_ref_table_formated.append(
                                    [q.id, fmt % f_format])
                        else:
                            data_ref_table_formated = data_ref_table

                    json_field['fields'] = self.db[ref_table].fields
                    
                    if json_field['editable']:
                        if 'reference_field' not in self.phanterpwa:
                            raise SyntaxError("".join(["You placed in '", self._field,
                                "' the reference field as editable, it is necessary to",
                                    " define the field of the referenced table. e.g. ",
                                        "phanterpwa = {'editable': True, 'refence_field': 'field_of_referenced_table'}"]))
                        else:
                            json_field['reference_field'] = self.phanterpwa["reference_field"]
                    else:
                        json_field['reference_field'] = 'id'
                    json_field['reference_table'] = ref_table
                if self._custom_label:
                    json_field['label'] = self._custom_label
                else:
                    json_field['label'] = FieldInst.label
                json_field['value'] = self.phanterpwa.get("value", default)
                json_field['type'] = 'reference'
                json_field['data_set'] = data_ref_table_formated
                if "type" in self.phanterpwa:
                    json_field["type"] = self.phanterpwa["type"]

            elif (t == "datetime") or\
                    (t == "date") or\
                    (t == "time"):
                default = FieldInst.default
                if self._record_id:
                    default = self._record[FieldInst]
                # print(type(default))
                # if default:
                #     default = default.isoformat()
                if "format" in self.phanterpwa:
                    dformat = self.phanterpwa["format"]
                elif t == "datetime":
                    dformat = "yyyy-MM-dd HH:mm:ss"
                elif t == "date":
                    dformat = "yyyy-MM-dd"
                elif t == "time":
                    dformat = "HH:mm:ss"
                if "validator_format" in self.phanterpwa and default:
                    formato_saida = self.phanterpwa["validator_format"]
                    json_field["validator_format"] = formato_saida
                    if t == "datetime":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%Y-%m-%d %H:%M:%S')
                        else:
                            tempo = time.strptime(str(default), '%Y-%m-%d %H:%M:%S')
                            resultado = str(time.strftime(formato_saida, tempo))
                    elif t == "date":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%Y-%m-%d')
                        else:
                            tempo = time.strptime(str(default), '%Y-%m-%d')
                            resultado = str(time.strftime(formato_saida, tempo))
                    elif t == "time":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%H:%M:%S')
                        else:
                            tempo = time.strptime(str(default), '%H:%M:%S')
                            resultado = str(time.strftime(formato_saida, tempo))
                    json_field['formatted'] = resultado
                elif default:
                    if t == "datetime":
                        general_format = '%Y-%m-%d %H:%M:%S'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(general_format, tempo))
                    elif t == "date":
                        general_format = '%Y-%m-%d'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(general_format, tempo))
                    elif t == "time":
                        general_format = '%H:%M:%S'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(formato_saida, tempo))
                    json_field['formatted'] = resultado
                else:
                    formato_saida = self.phanterpwa["validator_format"]
                    json_field["validator_format"] = formato_saida
                if self.simple_widget:
                    json_field['label'] = FieldInst.label
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = "string"
                    json_field['format'] = dformat
                    json_field['mask'] = str(dformat).replace(
                        "d", "#"
                    ).replace(
                        "M", "#"
                    ).replace(
                        "y", "#"
                    ).replace(
                        "H", "#"
                    ).replace(
                        "m", "#"
                    ).replace(
                        "s", "#"
                    )
                    new_validators = []
                    if "validators" in json_field and isinstance(json_field["validators"], list):
                        for x in json_field["validators"]:
                            if "IS_DATE" in x or "IS_TIME" in x:
                                new_validators.append("MASK:{0}".format(json_field['mask']))
                            else:
                                new_validators.append(x)
                        json_field["validators"] = new_validators
                else:
                    json_field['label'] = FieldInst.label
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = FieldInst.type
                    json_field['format'] = dformat
                    json_field['mask'] = dformat

            else:
                default = FieldInst.default
                if self._record_id:
                    default = self._record[FieldInst]
                if FieldInst.type == "id":
                    json_field['label'] = self.phanterpwa.get("label", FieldInst.label)
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = 'id'

                else:
                    json_field['label'] = self.phanterpwa.get("label", FieldInst.label)
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = FieldInst.type

                    if "format" in self.phanterpwa:
                        json_field["format"] = self.phanterpwa["format"]
                    if "data_set" in self.phanterpwa:
                        if FieldInst.type == "string":
                            json_field["type"] = "select"
                        if callable(self.phanterpwa['data_set']):
                            json_field["data_set"] = self.phanterpwa['data_set'](self.record)
                        elif isinstance(self.phanterpwa['data_set'], types.GeneratorType):
                            json_field["data_set"] = list(self.phanterpwa['data_set'])
                            self.phanterpwa['data_set'] = json_field["data_set"]
                        else:
                            json_field["data_set"] = self.phanterpwa["data_set"]
                    if "mask" in self.phanterpwa:
                        json_field["mask"] = self.phanterpwa["mask"]
                    if "type" in self.phanterpwa:
                        json_field["type"] = self.phanterpwa["type"]
                    if "widget" in self.phanterpwa:
                        json_field["widget"] = self.phanterpwa["widget"]
                    if "cutter" in self.phanterpwa:
                        json_field["cutter"] = self.phanterpwa["cutter"]
                    if "scale-factor" in self.phanterpwa:
                        json_field["scale-factor"] = self.phanterpwa["scale-factor"]
                    if "width" in self.phanterpwa:
                        json_field["width"] = self.phanterpwa["width"]
                    if "height" in self.phanterpwa:
                        json_field["height"] = self.phanterpwa["height"]
                    if "view_width" in self.phanterpwa:
                        json_field["view_width"] = self.phanterpwa["view_width"]
                    if "view_height" in self.phanterpwa:
                        json_field["view_height"] = self.phanterpwa["view_height"]
            if "extra" in self.phanterpwa and self.record and callable(self.phanterpwa['extra']):
                json_field["extra"] = self.phanterpwa['extra'](self.record)
            if "url" in self.phanterpwa and self.record and callable(self.phanterpwa['url']):
                json_field["url"] = self.phanterpwa['url'](self.record)
            if "no-cache" in self.phanterpwa:
                json_field["no-cache"] = self.phanterpwa['no-cache']
            if "disabled" in self.phanterpwa:
                json_field["disabled"] = self.phanterpwa['disabled']
            if "editable" in self.phanterpwa:
                json_field["editable"] = self.phanterpwa['editable']
            if "wear" in self.phanterpwa:
                json_field["wear"] = self.phanterpwa['wear']
            if "ajax_data_set" in self.phanterpwa:
                json_field["ajax_data_set"] = self.phanterpwa['ajax_data_set']
            for x in self.phanterpwa:
                if x.startswith("_"):
                    json_field[x] = self.phanterpwa[x]
            if self._data_view:
                json_field["data_view"] = True
            else:
                json_field["data_view"] = False
        return json_field

    def as_xml(self):
        return None


class CustomField():
    def __init__(self, table, field, record_id=None, data_view=False):
        self.table = table
        self._db = self.table._db
        self._data_view = data_view
        self._field = field
        self._record = None
        self.record_id = record_id
        self.position = None
        self.section = None
        self.group = None
        self.out_of_form = False
        self.phanterpwa = dict()

        if hasattr(self._field, "phanterpwa"):
            for x in self._field.phanterpwa:
                if x not in self.phanterpwa:
                    self.phanterpwa[x] = self._field.phanterpwa[x]

            if "data_view" in self.phanterpwa:
                self._data_view = self.phanterpwa["data_view"]
            if "position" in self.phanterpwa:
                self.position = self.phanterpwa["position"]
            if "section" in self.phanterpwa:
                self.section = self.phanterpwa["section"]
            elif "group" in self.phanterpwa:
                self.group = self.phanterpwa["group"]
            if "out_of_form" in self.phanterpwa and\
                    self.phanterpwa["out_of_form"] is True:
                self.out_of_form = True
            if "simple_widget" in self.phanterpwa and\
                    self.phanterpwa["simple_widget"] is True:
                self.simple_widget = True
        if "_class" not in self.phanterpwa:
            self.phanterpwa["_class"] = "p-col w1p100"

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if value is None:
            self._position = None
        elif isinstance(value, int):
            self._position = value
        elif str(value).isdigit():
            self._position = int(value)
        else:
            raise SyntaxError(
                "The position value must be intenger. Give: {0}".format(type(value)))

    @property
    def record(self):
        return self._record

    @property
    def record_id(self):
        return self._record_id

    @record_id.setter
    def record_id(self, value):
        self._record_id = None
        if value:
            r = self._db(self._db[self.table._tablename]._id == value).select().first()
            if r:
                self._record = r
                self._record_id = r.id
            else:
                raise ValueError("The record_id not find in table")

    @property
    def db(self):
        return self._db

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        if isinstance(value, Table):
            self._table = value
        else:
            raise ValueError("The table must be a pydal Table instance.")

    def __ppwa_check_validators(self, validator):

        if isinstance(validator, IS_NOT_EMPTY):
            if "can_empty" not in self.phanterpwa:
                self.phanterpwa["can_empty"] = False
            return ["IS_NOT_EMPTY"]
        elif isinstance(validator, IS_DATETIME):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_DATETIME:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_DATE):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_DATE:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_TIME):
            if "format" not in self.phanterpwa:
                self.phanterpwa["validator_format"] = validator.format
                self.phanterpwa["format"] = datetime_formater(validator.format)
            return ["IS_TIME:{0}".format(datetime_formater(validator.format))]
        elif isinstance(validator, IS_ACTIVATION_CODE):
            return ["IS_ACTIVATION_CODE"]
        elif isinstance(validator, IS_EMAIL):
            return ["IS_EMAIL"]
        elif isinstance(validator, IS_IN_DB):
            validator.build_set()
            table_name = validator.ktable
            if callable(self._db[table_name]._format):
                theset = [self._db[table_name]._format(x) for x in self._db(self._db[table_name].id > 0).select()]
            elif self._db[table_name]._format and isinstance(self._db[table_name]._format, str):
                theset = [self._db[table_name]._format % x.as_dict() for x in self._db(self._db[table_name].id > 0).select()]
            else:
                theset = [x.id for x in self._db(self._db[table_name].id > 0).select()]
            return ["IS_IN_SET:{0}".format(json.dumps("{0}".format(theset)))]
        elif isinstance(validator, IS_IN_SET):
            if "data_set" not in self.phanterpwa and "type" not in self.phanterpwa:
                self.phanterpwa["type"] = "select"
                self.phanterpwa["data_set"] = list(validator.theset)
            return ["IS_IN_SET:{0}".format(json.dumps("{0}".format(validator.theset)))]
        elif isinstance(validator, IS_EMPTY_OR):
            if "can_empty" not in self.phanterpwa and \
                    "validators" not in self.phanterpwa:
                self.phanterpwa["can_empty"] = True
            elif "validators" in self.phanterpwa and \
                    "IS_NOT_EMPTY" in self.phanterpwa["validators"]:
                self.phanterpwa["can_empty"] = False
            other = self.__ppwa_check_validators(validator.other)
            if other:
                if isinstance(other, (list, tuple)):
                    return ["IS_EMPTY_OR", *other]
                else:
                    return ["IS_EMPTY_OR", other]
            else:
                return []
        elif isinstance(validator, type(lambda: 0)):
            return []
        elif isinstance(validator, list):
            return [self.__ppwa_check_validators(x) for x in validator]

    def as_dict(self):
        json_field = dict()
        if not self.out_of_form:
            FieldInst = self._field
            t = FieldInst.type
            def u_list(v_list):
                result = []
                if isinstance(v_list, list):
                    t = list(v_list)
                    for item in v_list:
                        if (isinstance(item, (list, tuple))):
                            result.extend(item)
                        elif item:
                            result.append(item)
                return result
            auto_validators = u_list(self.__ppwa_check_validators(FieldInst.requires))

            if "validators" not in self.phanterpwa:
                json_field["validators"] = auto_validators
            else:
                json_field["validators"] = self.phanterpwa["validators"]
            default = FieldInst.default
            if "can_empty" in self.phanterpwa:
                json_field["can_empty"] = self.phanterpwa["can_empty"]

            if t.startswith("reference"):

                if self._record_id:
                    if self._record[FieldInst]:
                        default = self._record[FieldInst].id
                    else:
                        default = None
                if not self.db:
                    raise SyntaxError("The db must be DAL database instance")
                ref_table = t.split(" ")[1]
                ref_fields = []
                for f in self.db[ref_table].fields:
                    ref_fields.append(f)
                data_ref_table = []
                data_ref_table_formated = []
                q_ref_table = self.db(self.db[ref_table]).select()
                fmt = None
                if self.db[ref_table]._format:
                    fmt = self.db[ref_table]._format
                for q in q_ref_table:
                    f_format = {}
                    row = []
                    for z in ref_fields:
                        if (self.db[ref_table][z].type == "datetime") or\
                                (self.db[ref_table][z].type == "date") or\
                                (self.db[ref_table][z].type == "time"):
                            if q[z]:
                                row.append(q[z].isoformat())
                            else:
                                row.append(q[z])
                        else:
                            row.append(q[z])
                        f_format[z] = q[z]
                    data_ref_table.append(row)
                    if fmt:
                        if hasattr(fmt, '__call__'):
                            data_ref_table_formated.append(
                                [q.id, fmt(q) % f_format])
                        else:
                            data_ref_table_formated.append(
                                [q.id, fmt % f_format])
                    else:
                        data_ref_table_formated = data_ref_table

                json_field['label'] = FieldInst.label
                json_field['value'] = self.phanterpwa.get("value", default)
                json_field['type'] = 'reference'
                json_field['fields'] = self.db[ref_table].fields
                json_field['data_set'] = data_ref_table_formated
                json_field['editable'] = self.phanterpwa.get("editable", False)
                if "type" in self.phanterpwa:
                    json_field["type"] = self.phanterpwa["type"]
                if json_field['editable']:
                    if 'reference_field' not in self.phanterpwa:
                        raise SyntaxError("".join(["You placed in '", self._field,
                            "' the reference field as editable, it is necessary to",
                                " define the field of the referenced table. e.g. ",
                                    "phanterpwa = {'editable': True, 'refence_field': 'field_of_referenced_table'}"]))
                    else:
                        json_field['reference_field'] = self.phanterpwa["reference_field"]
                else:
                    json_field['reference_field'] = 'id'
                json_field['reference_table'] = ref_table

            elif (t == "datetime") or\
                    (t == "date") or\
                    (t == "time"):
                default = FieldInst.default
                if self._record_id:
                    default = self._record[FieldInst]
                # print(type(default))
                # if default:
                #     default = default.isoformat()
                if "format" in self.phanterpwa:
                    dformat = self.phanterpwa["format"]
                elif t == "datetime":
                    dformat = "yyyy-MM-dd HH:mm:ss"
                elif t == "date":
                    dformat = "yyyy-MM-dd"
                elif t == "time":
                    dformat = "HH:mm:ss"
                if "validator_format" in self.phanterpwa and default:
                    formato_saida = self.phanterpwa["validator_format"]
                    json_field["validator_format"] = formato_saida
                    if t == "datetime":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%Y-%m-%d %H:%M:%S')
                        else:
                            tempo = time.strptime(str(default), '%Y-%m-%d %H:%M:%S')
                            resultado = str(time.strftime(formato_saida, tempo))
                    elif t == "date":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%Y-%m-%d')
                        else:
                            tempo = time.strptime(str(default), '%Y-%m-%d')
                            resultado = str(time.strftime(formato_saida, tempo))
                    elif t == "time":
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(formato_saida))
                            default = default.strftime('%H:%M:%S')
                        else:
                            tempo = time.strptime(str(default), '%H:%M:%S')
                            resultado = str(time.strftime(formato_saida, tempo))
                    json_field['formatted'] = resultado
                elif default:
                    if t == "datetime":
                        general_format = '%Y-%m-%d %H:%M:%S'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(general_format, tempo))
                    elif t == "date":
                        general_format = '%Y-%m-%d'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(general_format, tempo))
                    elif t == "time":
                        general_format = '%H:%M:%S'
                        json_field["validator_format"] = general_format
                        if isinstance(default, datetime.datetime) or isinstance(default, datetime.date):
                            resultado = str(default.strftime(general_format))
                            default = resultado
                        else:
                            tempo = time.strptime(str(default), general_format)
                            resultado = str(time.strftime(formato_saida, tempo))
                    json_field['formatted'] = resultado
                else:
                    formato_saida = self.phanterpwa["validator_format"]
                    json_field["validator_format"] = formato_saida

                json_field['label'] = FieldInst.label
                json_field['value'] = self.phanterpwa.get("value", default)
                json_field['type'] = FieldInst.type
                json_field['format'] = dformat
                json_field['mask'] = dformat

            else:
                default = FieldInst.default
                if FieldInst.type == "id":
                    json_field['label'] = FieldInst.label
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = 'id'

                else:
                    json_field['label'] = FieldInst.label
                    json_field['value'] = self.phanterpwa.get("value", default)
                    json_field['type'] = FieldInst.type

                    if "format" in self.phanterpwa:
                        json_field["format"] = self.phanterpwa["format"]
                    if "data_set" in self.phanterpwa:
                        if FieldInst.type == "string":
                            json_field["type"] = "select"
                        if callable(self.phanterpwa['data_set']):
                            json_field["data_set"] = self.phanterpwa['data_set'](self.record)
                        else:
                            json_field["data_set"] = self.phanterpwa["data_set"]
                    if "mask" in self.phanterpwa:
                        json_field["mask"] = self.phanterpwa["mask"]
                    if "type" in self.phanterpwa:
                        json_field["type"] = self.phanterpwa["type"]
                    if "cutter" in self.phanterpwa:
                        json_field["cutter"] = self.phanterpwa["cutter"]
                    if "scale-factor" in self.phanterpwa:
                        json_field["scale-factor"] = self.phanterpwa["scale-factor"]
                    if "width" in self.phanterpwa:
                        json_field["width"] = self.phanterpwa["width"]
                    if "height" in self.phanterpwa:
                        json_field["height"] = self.phanterpwa["height"]
                    if "view_width" in self.phanterpwa:
                        json_field["view_width"] = self.phanterpwa["view_width"]
                    if "view_height" in self.phanterpwa:
                        json_field["view_height"] = self.phanterpwa["view_height"]
            if "extra" in self.phanterpwa and self.record and callable(self.phanterpwa['extra']):
                json_field["extra"] = self.phanterpwa['extra'](self.record)
            if "url" in self.phanterpwa and self.record and callable(self.phanterpwa['url']):
                json_field["url"] = self.phanterpwa['url'](self.record)

            if "no-cache" in self.phanterpwa:
                json_field["no-cache"] = self.phanterpwa['no-cache']
            if "disabled" in self.phanterpwa:
                json_field["disabled"] = self.phanterpwa['disabled']
            if "editable" in self.phanterpwa:
                json_field["editable"] = self.phanterpwa['editable']
            if "ajax_data_set" in self.phanterpwa:
                json_field["ajax_data_set"] = self.phanterpwa['ajax_data_set']
            for x in self.phanterpwa:
                if x.startswith("_"):
                    json_field[x] = self.phanterpwa[x]
            if self._data_view:
                json_field["data_view"] = True
            else:
                json_field["data_view"] = False
        return json_field

    def as_xml(self):
        return None


class FormFromTableDAL():
    def __init__(self, table, record_id=None, fields=None, custom_validates={}, data_view=False, simple_form=False):
        self.table = table
        self._db = self.table._db
        self._data_view = data_view
        self.simple_form = simple_form
        if fields:
            self.fields = fields
        else:
            self._fields = self.table.fields
        self._record = None
        self._extra_fields = {}
        self.record_id = record_id
        self.json_widgets = []
        self.xml_wigets = {}
        self._widgets = {}
        self._errors = {}
        self._ok = {}
        self._verified = {}
        self._ignored = {}
        self._changed = {}
        self.custom_validates = custom_validates
        self._custom_phanterpwa = dict()

    @property
    def custom_phanterpwa(self):
        return self._custom_phanterpwa

    def add_extra_field(self, value):
        if isinstance(value, Field):
            if value.name in self.fields:
                raise ValueError("The name of extra field must be unique.")
            self._extra_fields[value.name] = value

    @property
    def record_id(self):
        return self._record_id

    @record_id.setter
    def record_id(self, value):
        self._record_id = None
        if value:
            r = self.table[value]
            if r:
                self._record = r
                self._record_id = value
            else:
                raise ValueError("The record_id not find in table")

    @property
    def record(self):
        return self._record

    @property
    def db(self):
        return self._db

    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, value):
        new_list = list()
        if isinstance(value, (list, tuple)):
            for l in value:
                if l in self.table.fields:
                    new_list.append(l)
            self._fields = new_list
        else:
            raise TypeError("The fields must be list or tuple. Given: {0}".format(type(value)))

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        if isinstance(value, Table):
            self._table = value
        else:
            raise ValueError("The table must be a pydal Table instance.")

    def validate(self, **dict_args):
        self._process()
        self._errors = {}
        self._verified = {}
        self._ok = {}
        self._ignored = {}
        for key in dict_args:
            new_value = dict_args[key]
            if key in self._widgets and key in self.fields:
                if key in self.custom_validates:
                    if callable(self.custom_validates[key]):
                        r = self.custom_validates[key](new_value)
                        if len(r) == 2 and isinstance(r, (tuple, list)):
                            if r[1] is not None:
                                self._errors[key] = r[1]
                            else:
                                self._ok[key] = dict_args[key]
                        else:
                            raise ValueError("The custom validate must return tuple or list with 2 values")
                    else:
                        raise ValueError("The custom_validate must be callable.")
                else:
                    w = self._widgets[key]
                    if ("type" in w) and w["type"] == "reference":
                        
                        t_name = w['reference_table']
                        f_name = w['reference_field']
                        if "editable" in w and w["editable"]:
                            if isinstance(new_value, str) and ref_values.match(new_value):
                                v_splited = "".join(new_value.split(":")[1:])
                                id_ = 0
                                q = self._db(self._db[t_name][f_name] == v_splited).select().first()
                                if q:
                                    id_ = q.id
                                else:
                                    id_ = self._db[t_name].insert(**{f_name: v_splited})
                                    q = self._db(self._db[t_name]['id'] == id_).select().first()
                                    q.update_record()
                                    self._db.commit()
                                if id_:
                                    dict_args[key] = id_
                                    self._ok[key] = dict_args[key]
                                else:
                                    me = w.get("message_error", "The new value for some reason cannot be added.")
                                    self._errors[key] = me
                            else:
                                if self.table[key].validate(new_value)[1] is not None:
                                    self._errors[key] = self.table[key].validate(new_value)[1]

                        else:
                            if self.table[key].validate(new_value)[1] is not None:
                                self._errors[key] = self.table[key].validate(new_value)[1]
                        if dict_args[key]:
                            self._verified[key] = int(dict_args[key])
                        else:
                            self._verified[key] = None
                    elif ("type" in w) and (w["type"] == "date" or w["type"] == "datetime" or w["type"] == "time"):
                        t = w["type"]
                        if dict_args[key]:
                            if t == "datetime":
                                validator_format = w.get("validator_format", '%Y-%m-%d %H:%M:%S')
                                try:
                                    tempo = time.strptime(str(dict_args[key]), validator_format)
                                    resultado = time.strftime('%Y-%m-%d %H:%M:%S', tempo)
                                    self._ok[key] = resultado
                                except ValueError:
                                    resultado = None
                                    self._errors[key] = self.table[key].validate(new_value)[1]
                            elif t == "date":
                                validator_format = w.get("validator_format", '%Y-%m-%d')
                                try:
                                    tempo = time.strptime(str(dict_args[key]), validator_format)
                                    resultado = time.strftime('%Y-%m-%d', tempo)
                                    self._ok[key] = resultado
                                except ValueError:
                                    resultado = None
                                    self._errors[key] = self.table[key].validate(new_value)[1]
                            elif t == "time":
                                validator_format = w.get("validator_format", '%H:%M:%S')
                                try:
                                    tempo = time.strptime(str(dict_args[key]), validator_format)
                                    resultado = time.strftime('%H:%M:%S', tempo)
                                    self._ok[key] = resultado
                                except ValueError:
                                    resultado = None
                                    self._errors[key] = self.table[key].validate(new_value)[1]
                            self._verified[key] = resultado
                        else:
                            self._verified[key] = None
                    elif ("type" in w) and (w["type"] == "boolean"):
                        if self.table[key].validate(new_value)[1] is not None:
                            self._errors[key] = self.table[key].validate(new_value)[1]
                        else:
                            self._ok[key] = dict_args[key]
                        if dict_args[key]:
                            self._verified[key] = True
                        elif dict_args[key] is False:
                            self._verified[key] = False
                        else:
                            self._verified[key] = None
                    else:
                        if self.table[key].validate(new_value)[1] is not None:
                            self._errors[key] = self.table[key].validate(new_value)[1]
                        else:
                            self._ok[key] = dict_args[key]
                        if dict_args[key]:
                            self._verified[key] = dict_args[key]
                        else:
                            self._verified[key] = None

            else:
                self._ignored[key] = dict_args[key]
        return self._errors

    def update_or_insert(self, **dict_args):
        self._changed = {}
        val = self.validate(**dict_args)
        if val:
            return val
        else:
            if self.record_id:
                r = self._db(self._db[self.table._tablename]._id == self.record_id).select(
                    ).first().update_record(**self._verified)
                self._record = r
            else:
                r = self.table.insert(**self._verified)
                self._record = self.table[r]
                self.record_id = r
            for key in self._verified:
                if key in self._widgets and key in self.fields:
                    if self._verified[key] and 'value' in self._widgets[key] and self._widgets[key]['value']:
                        if self._widgets[key]['type'] == "reference":
                            if int(self._verified[key]) != int(self._widgets[key]['value']):
                                self._changed[key] = [self._verified[key], self._widgets[key]['value']]
                        elif self._widgets[key]['type'] == 'date' or\
                                self._widgets[key]['type'] == 'datetime' or self._widgets[key]['type'] == 'time':
                            if self._verified[key] != self._widgets[key]['value']:
                                self._changed[key] = [self._verified[key], self._widgets[key]['value']]
                        elif str(self._verified[key]) != str(self._widgets[key]['value']):
                            self._changed[key] = [self._verified[key], self._widgets[key]['value']]
                    elif (self._verified[key] and 'value' in self._widgets[key]) or ('value' in self._widgets[key] and self._widgets[key]['value']):
                        self._changed[key] = [self._verified[key], self._widgets[key]['value']]
                    self._widgets[key]['value'] = self._verified[key]
            self._db.commit()
            self._process()

    def _process(self):
        self.dict = {
            "table": self.table._tablename,
            "id": self.record_id,
            "data_view": self._data_view,
            "widgets": self.json_widgets,
        }
        self._widgets = dict()
        section = dict()
        group = dict()
        for x in self.fields:
            if x in self.custom_phanterpwa:
                json_widget = WidgetFromFieldDALFromTableDAL(self.table, x, self.record_id, data_view=self._data_view, simple_widget=self.simple_form, custom_phanterpwa=self.custom_phanterpwa[x])
            else:
                json_widget = WidgetFromFieldDALFromTableDAL(self.table, x, self.record_id, data_view=self._data_view, simple_widget=self.simple_form)
            self._widgets[x] = json_widget.as_dict()
            if x != "id" or self.record_id:
                if json_widget.section:
                    if json_widget.section in section:
                        section[json_widget.section][1][1].append(["widget", [x, json_widget.as_dict()]])
                    else:
                        section[json_widget.section] = ["section",
                            [json_widget.section, [["widget", [x, json_widget.as_dict()]]]]]
                        self.json_widgets.append(section[json_widget.section])
                elif json_widget.group:
                    if json_widget.group in group:
                        group[json_widget.group][1][1].append(["widget", [x, json_widget.as_dict()]])
                    else:
                        group[json_widget.group] = ["group", [json_widget.group,
                            [["widget", [x, json_widget.as_dict()]]]]]
                        self.json_widgets.append(group[json_widget.group])
                elif json_widget.out_of_form:
                    pass
                else:
                    self.json_widgets.append(["widget", [x, json_widget.as_dict()]])
        for e in self._extra_fields:
            # self.json_widgets[e] = self._extra_fields[e].as_dict()
            # self._widgets[x] = self.json_widgets[e].as_dict()
            json_widget = CustomField(self.table, self._extra_fields[e], self.record_id, self._data_view)
            #WidgetFromFieldDALFromTableDAL(self.table, x, self.record_id, data_view=self._data_view)
            self._widgets[e] = json_widget.as_dict()
            if x != "id" or self.record_id:
                if json_widget.section:
                    if json_widget.section in section:
                        if json_widget.position or json_widget.position == 0:
                            section[json_widget.section][1][1].insert(json_widget.position, ["widget", [e, json_widget.as_dict()]])
                        else:
                            section[json_widget.section][1][1].append(["widget", [e, json_widget.as_dict()]])
                    else:
                        section[json_widget.section] = ["section",
                            [json_widget.section, [["widget", [e, json_widget.as_dict()]]]]]
                        self.json_widgets.append(section[json_widget.section])
                elif json_widget.group:
                    if json_widget.group in group:
                        if json_widget.position or json_widget.position == 0:
                            group[json_widget.group][1][1].insert(json_widget.position, ["widget", [e, json_widget.as_dict()]])
                        else:
                            group[json_widget.group][1][1].append(["widget", [e, json_widget.as_dict()]])
                    else:
                        group[json_widget.group] = ["group", [json_widget.group,
                            [["widget", [e, json_widget.as_dict()]]]]]
                        self.json_widgets.append(group[json_widget.group])
                elif json_widget.out_of_form:
                    pass
                else:
                    if json_widget.position or json_widget.position == 0:
                        self.json_widgets.insert(json_widget.position, ["widget", [e, json_widget.as_dict()]])
                    else:
                        self.json_widgets.append(["widget", [e, json_widget.as_dict()]])

    def as_dict(self):
        self._process()
        return self.dict

    def as_json(self, indent=2):
        return json.dumps(self.as_dict(), indent=indent)

    def as_xml(self, show_id=False, submit_button=None):
        self.show_id = show_id
        self._xml = FORM(
            DIV(
                DIV(
                    INPUT(**{
                        "_id": "phanterpwa-widget-input-input-{0}-csrf_token".format(self.table_name),
                        "_name": "csrf_token",
                        "_data-validators": ['IS_NOT_EMPTY'],
                        "_data-form": self.table_name,
                        "_type": "hidden"
                    }),
                ),
                _id='phanterpwa-widget-input-input-{0}-csrf_token'.format(self.table_name),
                _class='phanterpwa-widget phanterpwa-widget-hidden e-display_hidden'
            ),
            _id="phanterpwa-widget-form-{0}".format(self.table_name),
            _class="phanterpwa-widget phanterpwa-widget-form",
            _phanterpwa_jsonform=self.table_name
        )
        if submit_button is not None:
            self._buttons_container = DIV(
                submit_button,
                _class='phanterpwa-form-buttons-container'
            )
        else:
            self._buttons_container = DIV(
                DIV(
                    DIV(
                        DIV(
                            _id="phanterpwa-widget-form-submit_button-{0}".format(self.table_name),
                            _class="btn-autoresize wave_on_click waves-phanterpwa btn phanterpwa-widget-form-form_button link",
                            _phanterpwa_widget_submit_button=self.table_name
                        ),
                        _class="button-form"
                    ),
                    _class="phanterpwa-widget-form-form_button-container"
                ),
                _class='phanterpwa-form-buttons-container'
            )


class FieldsDALValidateDictArgs(object):

    def __init__(self, dict_args, *fields):
        super(FieldsDALValidateDictArgs, self).__init__()
        self.dict_args = dict_args
        self.fields = fields
        self._errors = {}
        self._errors_on_insert = {}
        self._errors_on_update = {}
        self._verified = {}
        self._ignored = {}

    @property
    def errors(self):
        self.validate()
        return self._errors

    @property
    def dict_args(self):
        return self._dict_args

    @dict_args.setter
    def dict_args(self, dict_args):
        if isinstance(dict_args, dict):
            self._dict_args = dict_args
        else:
            raise TypeError(
                "The dict_args must be dict object. given: %s" % type(dict_args)
            )

    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, list_fields):
        n_l = list()
        for x in list_fields:
            if isinstance(x, Field):
                n_l.append(x)
            else:
                if x is not None:
                    raise TypeError(
                        "The fields must be pydal.Field object. given: %s" % type(x)
                    )
        self._fields = n_l

    @property
    def verified(self):
        self.validate()
        return self._verified

    @property
    def ignored(self):
        self.validate()
        return self._ignored

    def validate(self):
        self._errors = {}
        self._verified = {}
        self._ignored = {}
        for f in self.dict_args:
            ig = True
            for F in self.fields:
                if F.name == f:
                    ig = False
                    result = F.validate(self.dict_args[f])
                    if result[1] is not None:
                        self._errors[f] = result[1]
                    else:
                        self._verified[f] = self.dict_args[f]
            if ig:
                self._ignored[f] = self.dict_args[f]
        if self._errors_on_insert:
            for x in self._errors_on_insert:
                self._errors[x] = self._errors_on_insert[x]
        if self._errors_on_update:
            for x in self._errors_on_update:
                self._errors[x] = self._errors_on_update[x]
        if self._errors:
            return self._errors

    def insert(self, dbtable, commit=True):
        self._errors_on_insert = {}
        self._errors_on_update = {}
        id_record = dbtable.insert(
            **self.dict_args
        )
        if id_record and commit:
            dbtable._db.commit()
            return id_record
        else:
            return None

    def update(self, dbtable, row_id, commit=True):
        self._errors_on_update = {}
        self._errors_on_insert = {}
        id_record = dbtable[row_id].update_record(
            **self.dict_args
        )
        if id_record and commit:
            dbtable._db.commit()
            return id_record
        else:
            return None

    def validate_and_insert(self, dbtable, commit=True):
        self._errors_on_insert = {}
        self._errors_on_update = {}
        self.validate()
        if not self.errors and isinstance(dbtable, Table):
            _local_fields = {}
            for f in dbtable.fields:
                veri = self.verified.get(f)
                if veri:
                    _local_fields[f] = veri
            if _local_fields:
                rep = dbtable.validate_and_insert(**_local_fields)
                if isinstance(rep, dict):
                    if rep.get("errors"):
                        self._errors_on_insert = rep.get("errors")
                        dbtable._db.rollback()
                    elif rep.get("id") and commit:
                        dbtable._db.commit()
                else:
                    if rep.errors:
                        self._errors_on_insert = dict(rep.errors)
                        dbtable._db.rollback()
                    elif rep.id and commit:
                        dbtable._db.commit()
                    return rep
        else:
            if not self.errors:
                raise "The dbtable must be pydal.DAL.Table instance. given: {0}.".format(type(dbtable))

    def validate_and_update(self, dbset, commit=True):
        self._errors_on_insert = {}
        self._errors_on_update = {}
        self.validate()
        if not self.errors and isinstance(dbset, Set):
            rep = dbset.validate_and_update(**self.verified)
            if isinstance(rep, dict):
                if rep.get("errors"):
                    self._errors_on_update = rep.get("errors")
                    dbset._db.rollback()
                elif rep.get("id") and commit:
                    dbset._db.commit()
            else:
                if rep.errors:
                    self._errors_on_update = dict(rep.errors)
                    dbset._db.rollback()
                elif rep.id and commit:
                    dbset._db.commit()
                return rep
        else:
            if not self.errors:
                raise "The dbtable must be pydal.Objects.Set instance. given: {0}.".format(type(dbset))



import unittest
import os
import glob
from phanterpwa.i18n import (
    Translator
)

sample1 = """[
  "hello world!"
]"""

sample2 = """{
  "hello world!": "hello world!"
}"""

sample4 = """{
  "abacaxi": "abacaxi",
  "hello world!": "hello world!"
}"""

sample5 = """{
  "abacaxi": "pineapple",
  "hello world!": "hello world!"
}"""

CURRENT_PATH = os.path.dirname(__file__)


class TestI18n(unittest.TestCase):
    def test1_created_file(self):
        files = glob.glob(os.path.join(CURRENT_PATH, "test_i18n_languages", "*.json"))
        for x in files:
            os.remove(x)
        Trans = Translator(os.path.join(CURRENT_PATH, "test_i18n_languages"), "test")
        self.assertTrue(os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages")))
        self.assertTrue(os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages", "entries.json")))
        Trans.add_language("pt-BR")
        self.assertTrue(os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-BR.json")))

    def test2_persistence(self):
        if os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages", "entries.json")):
            os.remove(os.path.join(CURRENT_PATH, "test_i18n_languages", "entries.json"))
        # Test create a new entries.json
        Trans = Translator(os.path.join(CURRENT_PATH, "test_i18n_languages"), "test")
        if os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-PT.json")):
            os.remove(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-PT.json"))
        # Test create a new pt-PT.json
        Trans.add_language("pt-PT")
        # Test add entrie without save
        Trans.T("hello world!")
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "entries.json"), "r") as f:
            self.assertEqual(f.readlines(), ['[]'])
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-PT.json"), "r") as f:
            self.assertEqual(f.readlines(), ['{}'])
        Trans.save()
        # Test entrie after save
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "entries.json"), "r") as f:
            self.assertEqual(f.read(), sample1)
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-PT.json"), "r") as f:
            self.assertEqual(f.read(), sample2)

    def test3_attr_translate(self):
        if os.path.exists(os.path.join(CURRENT_PATH, "test_i18n_languages", "en-US.json")):
            os.remove(os.path.join(CURRENT_PATH, "test_i18n_languages", "en-US.json"))
        Trans = Translator(os.path.join(CURRENT_PATH, "test_i18n_languages"), "test")
        # Translate just en-US language
        Trans.translate("en-US", "abacaxi", "pineapple")
        Trans.save()
        # abacaxi translate is abacaxi
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-BR.json"), "r") as f:
            self.assertEqual(f.read(), sample4)
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "pt-PT.json"), "r") as f:
            self.assertEqual(f.read(), sample4)
        # abacaxi translate is pineapple
        with open(os.path.join(CURRENT_PATH, "test_i18n_languages", "en-US.json"), "r") as f:
            self.assertEqual(f.read(), sample5)

    def test4_attr_translator(self):
        Trans = Translator(os.path.join(CURRENT_PATH, "test_i18n_languages"), "test")
        self.assertEqual("dont exists", Trans.translator("dont exists", "pt-BR"))
        self.assertEqual("abacaxi", Trans.translator("abacaxi", "pt-BR"))
        self.assertEqual("pineapple", Trans.translator("abacaxi", "en-US"))


if __name__ == '__main__':
    unittest.main()



import sqlite3
from datetime import (
    datetime,
    timedelta
)
from phanterpwa.backend.decorators import (
    check_application,
    check_client_token,
    check_user_token,
    async_retry_on_locked
)
from phanterpwa.i18n import browser_language
from phanterpwa.third_parties.xss import xssescape as E
from phanterpwa.captchasvg.captcha import Captcha
from phanterpwa.gallery.integrationDAL import PhanterpwaGalleryUserImage

from tornado import (
    web
)
from phanterpwa.backend.security import (
    Serialize,
    SignatureExpired,
    BadSignature,
    URLSafeSerializer,
)

from datetime import (
    datetime,
    timedelta
)

CLEANUP_TIME = datetime.now()

class SignClient(web.RequestHandler):
    """
        /api/client/
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))
        if self.i18nTranslator:
            self.i18nTranslator.direct_translation = self.phanterpwa_language
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    @check_application()
    async def get(self, *args, **kargs):
        expire_date = datetime.now() - timedelta(seconds=self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire'])
        anonymous_expire_date = datetime.now() - timedelta(seconds=self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire'])
        deactived_user_expire_date = datetime.now() - timedelta(seconds=self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me'])
        str_expire_date = expire_date.strftime('%Y-%m-%d %H:%M:%S')
        str_anonymous_expire_date = anonymous_expire_date.strftime('%Y-%m-%d %H:%M:%S')
        global CLEANUP_TIME
        if CLEANUP_TIME < datetime.now() - timedelta(seconds=600):
            CLEANUP_TIME = datetime.now()
            self.DALDatabase(
                (self.DALDatabase.client.date_created < str_expire_date)
            ).delete()
            self.DALDatabase(
                (self.DALDatabase.client.auth_user == None)
                & (self.DALDatabase.client.date_created < str_anonymous_expire_date)
            ).delete()
            self.DALDatabase(
                (self.DALDatabase.auth_user.activated == False)
                & ((self.DALDatabase.auth_user.date_created < deactived_user_expire_date) | (self.DALDatabase.auth_user.date_created == None))
            ).delete()

        self.phanterpwa_client_token = self.request.headers.get('phanterpwa-client-token')
        self.phanterpwa_authorization = self.request.headers.get('phanterpwa-authorization')
        t_client = Serialize(
            self.projectConfig['BACKEND'][self.app_name]['secret_key'],
            self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
        )
        t_url = URLSafeSerializer(
            self.projectConfig['BACKEND'][self.app_name]['secret_key'],
            salt="url_secret_key"
        )
        msg = 'The client does not have a client-token, a new one has been generated.'
        q = None
        if self.phanterpwa_client_token:
            q = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
            if q:
                token_content_client = None
                try:
                    token_content_client = t_client.loads(self.phanterpwa_client_token)
                except BadSignature:
                    msg = 'The client have a invalid client-token, a new one has been generated.'
                    token_content_client = None
                except SignatureExpired:
                    msg = 'The client have a expired client-token, a new one has been generated.'
                    token_content_client = None
                if token_content_client:
                    if token_content_client['user_agent'] == self.phanterpwa_user_agent:
                        if 'id_user' in token_content_client:
                            token_url = t_url.dumps({**token_content_client})
                            if self.phanterpwa_authorization:
                                t_user = Serialize(
                                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                    self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                                )
                                token_content_user = None
                                try:
                                    token_content_user = t_user.loads(self.phanterpwa_authorization)
                                except BadSignature:
                                    token_content_user = None
                                except SignatureExpired:
                                    token_content_user = None
                                if token_content_user and 'id' in token_content_user:
                                    id_user = token_content_user['id']
                                    q_user = self.DALDatabase(self.DALDatabase.auth_user.id == id_user).select().first()
                                    q_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == id_user) &
                                        (self.DALDatabase.client.token == self.phanterpwa_client_token)
                                    ).select().first()
                                    if q_user and q_client:
                                        msg = "".join(['The client-token is valid and belongs ',
                                            'to a login user, will be reused.'])
                                        time_resing_client = int(
                                            self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire'] * 0.8) + 1
                                        if q_client.date_created +\
                                                timedelta(seconds=time_resing_client) < datetime.now():
                                            new_client = self.DALDatabase.client.insert(
                                                auth_user=q_client.auth_user,
                                                remember_me=q_client.remember_me,
                                                last_resign=datetime.now(),
                                                date_created=datetime.now()
                                            )
                                            content = {
                                                'id_user': str(q_user.id),
                                                "id_client": str(new_client),
                                                'user_agent': self.phanterpwa_user_agent,
                                                'remote_addr': self.phanterpwa_remote_ip,
                                            }
                                            token_url = t_url.dumps(content)
                                            token_client = t_client.dumps(content)
                                            q_client.delete_record()
                                            q_client = self.DALDatabase(
                                                self.DALDatabase.client.id == new_client).select().first()
                                            q_client.update_record(
                                                token=token_client
                                            )
                                            self.phanterpwa_client_token = token_client
                                            msg = "".join(['The client-token is valid and belongs ',
                                                'to a login user, but a new one has been generated.'])
                                        if not q_user.permit_mult_login:
                                            self.DALDatabase(
                                                (self.DALDatabase.client.auth_user == id_user) &
                                                (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                            ).delete()
                                        q_role = self.DALDatabase(
                                            (self.DALDatabase.auth_membership.auth_user == q_user.id) &
                                            (self.DALDatabase.auth_group.id ==
                                                self.DALDatabase.auth_membership.auth_group)
                                        ).select(
                                            self.DALDatabase.auth_group.id,
                                            self.DALDatabase.auth_group.role,
                                            orderby=self.DALDatabase.auth_group.grade
                                        )
                                        roles = [x.role for x in q_role]
                                        dict_roles = {x.id: x.role for x in q_role}
                                        role = None
                                        if roles:
                                            role = roles[-1]
                                        self.DALDatabase.commit()
                                        user_image = PhanterpwaGalleryUserImage(
                                            q_user.id, self.DALDatabase, self.projectConfig)
                                        self.set_status(200)
                                        return self.write({
                                            'status': 'OK',
                                            'code': 200,
                                            'message': msg,
                                            'auth_user': {
                                                'id': q_user.id,
                                                'first_name': E(q_user.first_name),
                                                'last_name': E(q_user.last_name),
                                                'email': q_user.email,
                                                'remember_me': q_client.remember_me,
                                                'roles': roles,
                                                'role': role,
                                                'dict_roles': dict_roles,
                                                'activated': q_user.activated,
                                                'image': user_image.id_image,
                                                'two_factor': q_user.two_factor_login,
                                                'multiple_login': q_user.permit_mult_login,
                                                'locale': q_user.locale,
                                                'social_login': None
                                            },
                                            'i18n': {
                                                'message': self.T(msg),
                                                'auth_user': {
                                                    'role': self.T(role)
                                                }
                                            },
                                            'client_token': self.phanterpwa_client_token,
                                            'url_token': token_url,
                                        })
                            else:
                                msg = "".join(['The client-token is valid and belongs ',
                                    'to a logout user, will be reused.'])
                                self.set_status(200)
                                return self.write({
                                    'status': 'OK',
                                    'code': 200,
                                    'message': msg,
                                    'auth_user': 'logout',
                                    'i18n': {
                                        'message': self.T(msg),
                                        'auth_user': self.T('logout')
                                    },
                                    'client_token': self.phanterpwa_client_token,
                                    'url_token': token_url,
                                })
                        else:
                            msg = 'The client-token is valid and anonymous, will be reused'
                            self.set_status(200)
                            return self.write({
                                'status': 'OK',
                                'code': 200,
                                'message': msg,
                                'auth_user': 'anonymous',
                                'i18n': {
                                    'message': self.T(msg),
                                    'auth_user': self.T('anonymous')
                                },
                                'client_token': self.phanterpwa_client_token
                            })
                    else:
                        msg = "".join(['The client-token is valid but appears to ',
                            'belong to another client, a new one has been generated.'])
                else:
                    if self.phanterpwa_authorization:
                        t_user = Serialize(
                            self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                            self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                        )
                        token_content_user = None
                        try:
                            token_content_user = t_user.loads(self.phanterpwa_authorization)
                        except BadSignature:
                            token_content_user = None
                        except SignatureExpired:
                            token_content_user = None
                        if token_content_user and 'id' in token_content_user:
                            id_user = token_content_user['id']
                            if int(token_content_user['id']) == int(id_user):
                                q_user = self.DALDatabase(self.DALDatabase.auth_user.id == id_user).select().first()
                                if q_user:
                                    content = {
                                        'id_user': str(q_user.id),
                                        "id_client": str(q.id),
                                        'user_agent': self.phanterpwa_user_agent,
                                        'remote_addr': self.phanterpwa_remote_ip,
                                    }
                                    token_url = t_url.dumps(content)
                                    token_client = t_client.dumps(content)
                                    q.update_record(
                                        token=token_client,
                                        id_user=q_user.id,
                                        date_created=datetime.now()
                                    )
                                    if not q_user.permit_mult_login:
                                        self.DALDatabase(
                                            (self.DALDatabase.client.auth_user == id_user) &
                                            (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                        ).delete()
                                    q_role = self.DALDatabase(
                                        (self.DALDatabase.auth_membership.auth_user == q_user.id) &
                                        (self.DALDatabase.auth_group.id == self.DALDatabase.auth_membership.auth_group)
                                    ).select(
                                        self.DALDatabase.auth_group.id,
                                        self.DALDatabase.auth_group.role,
                                        orderby=self.DALDatabase.auth_group.grade
                                    )
                                    roles = [x.role for x in q_role]
                                    dict_roles = {x.id: x.role for x in q_role}
                                    role = None
                                    if roles:
                                        role = roles[-1]
                                    self.DALDatabase.commit()
                                    msg = "".join(['The client token is valid but expired, ',
                                        'a new one has been generated.'])
                                    user_image = PhanterpwaGalleryUserImage(
                                        q_user.id, self.DALDatabase, self.projectConfig)
                                    self.set_status(200)
                                    return self.write({
                                        'status': 'OK',
                                        'code': 200,
                                        'message': msg,
                                        'auth_user': {
                                            'id': q_user.id,
                                            'first_name': E(q_user.first_name),
                                            'last_name': E(q_user.last_name),
                                            'email': q_user.email,
                                            'remember_me': q.remember_me,
                                            'roles': roles,
                                            'role': role,
                                            'dict_roles': dict_roles,
                                            'activated': q_user.activated,
                                            'image': user_image.id_image,
                                            'two_factor': q_user.two_factor_login,
                                            'multiple_login': q_user.permit_mult_login,
                                            'locale': q_user.locale,
                                            'social_login': None
                                        },
                                        'i18n': {
                                            'message': self.T(msg),
                                            'auth_user': {
                                                'role': self.T(role)
                                            }
                                        },
                                        'client_token': token_client,
                                        'url_token': token_url
                                    })
                    else:
                        msg = 'The client have a invalid client-token, a new one has been generated.'
                        q.delete_record()
            else:
                msg = 'The client have a invalid client-token, not found, a new one has been generated.'

        if q:
            id_client = q.id
        else:
            id_client = self.DALDatabase.client.insert(date_created=datetime.now())

        content = {
            "id_client": id_client,
            'user_agent': self.phanterpwa_user_agent,
            'remote_addr': self.phanterpwa_remote_ip
        }
        token_client = t_client.dumps(content)
        q_client = self.DALDatabase(self.DALDatabase.client.id == id_client).select(self.DALDatabase.client.id).first()
        if q_client:
            q_client.update_record(token=token_client)
        self.DALDatabase.commit()
        self.set_status(200)
        return self.write({
            'status': 'OK',
            'code': 200,
            'message': msg,
            'auth_user': 'anonymous',
            'i18n': {
                'message': self.T(msg),
                'auth_user': self.T('anonymous')
            },
            'client_token': token_client
        })

    def options(self, *args):
        self.set_status(200)
        return self.write({"status": "OK"})


class SignForms(web.RequestHandler):
    """
        url: '/api/signforms/<form_identify>'
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None, list_forms=[]):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))
        if self.i18nTranslator:
            self.i18nTranslator.direct_translation = self.phanterpwa_language
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip
        self.list_forms = [*{
            "phanterpwa-form-activation",
            "phanterpwa-form-profile",
            "phanterpwa-form-change_password",
            "phanterpwa-form-change_account",
            "phanterpwa-form-auth_user",
            "phanterpwa-form-auth_group",
            *list_forms
        }]

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    @check_user_token(ignore_activation=True)
    def get(self, *args, **kargs):
        """
        Receive request to create and response with a token csrf or captcha
        """
        form_identify = args[0]
        list_forms = self.list_forms

        id_client = int(self.phanterpwa_client_token_checked["id_client"])
        if form_identify in list_forms:
            expire_date = datetime.now() - timedelta(seconds=self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire'])
            str_expire_date = expire_date.strftime('%Y-%m-%d %H:%M:%S')
            self.DALDatabase(
                (self.DALDatabase.csrf.date_created < str_expire_date)
            ).delete()
            t = Serialize(
                "csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire']
            )
            id_csrf = self.DALDatabase.csrf.insert(
                form_identify=form_identify,
                user_agent=self.phanterpwa_user_agent,
                ip=self.phanterpwa_remote_ip,
                client=id_client
            )
            if id_csrf:
                sign_captha = t.dumps({
                    'id': str(id_csrf),
                    'form_identify': form_identify,
                    'user_agent': self.phanterpwa_user_agent,
                    'ip': self.phanterpwa_remote_ip,
                    'user': self.phanterpwa_current_user.id
                })
                q_csrf = self.DALDatabase(self.DALDatabase.csrf.id == id_csrf).select().first()
                q_csrf.update_record(token=sign_captha)
                msg = "Form signed"
                self.DALDatabase.commit()
                self.set_status(200)
                return self.write({
                    "status": "OK",
                    "code": 200,
                    "message": msg,
                    "i18n": {
                        "message": self.T(msg)
                    },
                    "csrf": sign_captha
                })
        self.DALDatabase.commit()
        msg = "The form can't sign"
        self.set_status(400)
        return self.write({
            "status": "Bad Request",
            "code": 400,
            "message": msg,
            "i18n": {
                "message": self.T(msg)
            }
        })

    def options(self, *args):
        self.set_status(200)
        self.write({"status": "OK"})


class SignLockForm(web.RequestHandler):
    """
        url: '/api/signlockform/'
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))
        if self.i18nTranslator:
            self.i18nTranslator.direct_translation = self.phanterpwa_language
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    @check_client_token()
    def get(self, *args, **kargs):
        """
        Receive request to create and response with a token csrf or captcha
        """
        form_identify = "user_locked"
        id_client = int(self.phanterpwa_client_token_checked["id_client"])
        if self.phanterpwa_current_client and self.phanterpwa_current_client.locked:
            self.DALDatabase((self.DALDatabase.csrf.client == id_client) &
                (self.DALDatabase.csrf.form_identify == form_identify)).delete()
            t = Serialize(
                "csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire']
            )
            id_csrf = self.DALDatabase.csrf.insert(
                form_identify=form_identify,
                user_agent=self.phanterpwa_user_agent,
                ip=self.phanterpwa_remote_ip,
                client=id_client
            )
            if id_csrf:
                sign_captha = t.dumps({
                    'id': str(id_csrf),
                    'form_identify': form_identify,
                    'user_agent': self.phanterpwa_user_agent,
                    'ip': self.phanterpwa_remote_ip
                })
                q_csrf = self.DALDatabase(self.DALDatabase.csrf.id == id_csrf).select().first()
                q_csrf.update_record(token=sign_captha)
            msg = "Form signed"
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                "code": 200,
                "message": msg,
                "i18n": {
                    "message": self.T(msg)
                },
                "csrf": sign_captha
            })
        else:
            self.DALDatabase.commit()
            msg = "The form can't sign, the user is not locked."
            self.set_status(400)
            return self.write({
                "status": "Bad Request",
                "code": 400,
                "message": msg,
                "i18n": {
                    "message": self.T(msg)
                }
            })

    def options(self, *args):
        self.set_status(200)
        self.write({"status": "OK"})


class SignCaptchaForms(web.RequestHandler):
    """
        url: '/api/signcaptchaforms/<form_identify>'
    """

    def initialize(self, app_name, projectConfig, DALDatabase, Translator_captcha, i18nTranslator=None, logger_api=None, list_forms=[]):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        self.Translator_captcha = Translator_captcha
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))
        if self.i18nTranslator:
            self.i18nTranslator.direct_translation = self.phanterpwa_language
        if self.Translator_captcha:
            self.Translator_captcha.direct_translation = self.phanterpwa_language
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip
        self.list_forms = [*{
            "phanterpwa-form-login",
            "phanterpwa-form-register",
            "phanterpwa-form-request_password",
            "phanterpwa-form-login401",
            *list_forms
        }]

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    @check_client_token()
    def get(self, *args, **kargs):
        """
        Receive request to create and response with a token csrf or captcha
        """

        form_identify = args[0]
        list_forms = self.list_forms
        id_client = int(self.phanterpwa_client_token_checked["id_client"])
        if form_identify in list_forms:
            self.DALDatabase((self.DALDatabase.captcha.client == id_client) &
                (self.DALDatabase.captcha.form_identify == form_identify)).delete()
            captcha = Captcha(
                form_identify,
                secret_key="csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                time_token_expire=self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire'],
                translator=self.Translator_captcha
            )
            html_captcha = captcha.html.xml()
            signature = captcha.signature
            self.DALDatabase.captcha.insert(
                token=signature,
                form_identify=form_identify,
                user_agent=self.phanterpwa_user_agent,
                ip=self.phanterpwa_remote_ip,
                date_created=datetime.now(),
                client=id_client
            )
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                "code": 200,
                "captcha": html_captcha,
                "signature": signature,
                "message": "Captcha send",
                "i18n": {
                    "message": self.T("Captcha send")
                }
            })
        else:
            self.DALDatabase.commit()
            self.set_status(400)
            return self.write({
                "status": "Bad Request",
                "code": 400,
                "message": "The form can't sign",
                "i18n": {
                    "message": self.T("The form can't sign")
                }
            })

    @check_client_token()
    def post(self, *args, **kargs):
        """
        Receive captcha user choice and response with csrf token
        """
        form_identify = args[0]
        list_forms = self.list_forms
        dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
        choice = dict_arguments['user_choice']
        sign = dict_arguments['signature']
        id_client = int(self.phanterpwa_client_token_checked["id_client"])
        q_captcha = self.DALDatabase(self.DALDatabase.captcha.token == sign).select().first()
        if q_captcha:
            self.DALDatabase((self.DALDatabase.captcha.client == id_client) &
                (self.DALDatabase.captcha.form_identify == form_identify)).delete()
            self.DALDatabase((self.DALDatabase.csrf.client == id_client) &
                (self.DALDatabase.csrf.form_identify == form_identify)).delete()
            q_captcha.delete_record()
            captcha = Captcha(
                form_identify,
                "csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire'],
                translator=self.Translator_captcha
            )

            if captcha.check(sign, choice):
                id_csrf = self.DALDatabase.csrf.insert(
                    form_identify=form_identify,
                    user_agent=self.phanterpwa_user_agent,
                    ip=self.phanterpwa_remote_ip,
                    client=id_client
                )
                if id_csrf:
                    t = Serialize(
                        "csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                        self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire']
                    )
                    sign_captha = t.dumps({
                        'id': str(id_csrf),
                        'form_identify': form_identify,
                        'user_agent': self.phanterpwa_user_agent,
                        'ip': self.phanterpwa_remote_ip
                    })
                    q_csrf = self.DALDatabase(self.DALDatabase.csrf.id == id_csrf).select().first()
                    q_csrf.update_record(token=sign_captha)
                self.DALDatabase.commit()
                self.set_status(200)
                return self.write({
                    "status": "OK",
                    "code": 200,
                    "message": "Captcha resolved",
                    "i18n": {
                        "message": self.T("Captcha resolved")
                    },
                    "captcha": captcha.html_ok.xml(),
                    "csrf": sign_captha
                })
            else:

                id_client = int(self.phanterpwa_client_token_checked["id_client"])
                if form_identify in list_forms:
                    self.DALDatabase((self.DALDatabase.captcha.client == id_client) &
                        (self.DALDatabase.captcha.form_identify == form_identify)).delete()
                    captcha = Captcha(
                        form_identify,
                        secret_key="csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                        time_token_expire=self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire'],
                        translator=self.Translator_captcha
                    )
                    html_captcha = captcha.html.xml()
                    signature = captcha.signature
                    self.DALDatabase.captcha.insert(
                        token=signature,
                        form_identify=form_identify,
                        user_agent=self.phanterpwa_user_agent,
                        ip=self.phanterpwa_remote_ip,
                        date_created=datetime.now(),
                        client=id_client
                    )
                    self.DALDatabase.commit()
                    self.set_status(400)
                    return self.write({
                        "status": "Bad Request",
                        "code": 400,
                        "captcha": html_captcha,
                        "signature": signature,
                        "message": "Incorrect captcha",
                        "i18n": {
                            "message": self.T("Incorrect captcha")
                        }
                    })
                else:
                    self.DALDatabase.commit()
                    self.set_status(400)
                    return self.write({
                        "status": "Bad Request",
                        "code": 400,
                        "message": "Incorrect captcha",
                        "i18n": {
                            "message": self.T("Incorrect captcha")
                        }
                    })
        else:
            self.DALDatabase.commit()
            self.set_status(400)
            return self.write({
                "status": "Bad Request",
                "code": 400,
                "message": "Incorrect captcha, not found.",
                "i18n": {
                    "message": self.T("Incorrect captcha, not found.")
                }
            })

    def options(self, *args, **kargs):
        self.set_status(200)
        return self.write({"status": "OK"})


class ReSing(web.RequestHandler):
    """
        url: '/api/resigncredentials/''
    """

    def initialize(self, app_name, projectConfig, DALDatabase, i18nTranslator=None, logger_api=None):
        self.app_name = app_name
        self.projectConfig = projectConfig
        self.DALDatabase = DALDatabase
        self.i18nTranslator = i18nTranslator
        if logger_api:
            self.logger_api = logger_api
        if i18nTranslator:
            self.T = i18nTranslator.T
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header(
            "Access-Control-Allow-Headers",
            "".join([
                "phanterpwa-language,",
                "phanterpwa-application,",
                "phanterpwa-application-version,",
                "phanterpwa-client-token,",
                "phanterpwa-authorization,",
                "cache-control"
            ])
        )
        self.set_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
        if self.request.headers.get("phanterpwa-language"):
            self.phanterpwa_language = self.request.headers.get("phanterpwa-language")
        else:
            self.phanterpwa_language = browser_language(self.request.headers.get("Accept-Language"))
        if self.i18nTranslator:
            self.i18nTranslator.direct_translation = self.phanterpwa_language
        self.phanterpwa_user_agent = str(self.request.headers.get('User-Agent'))
        self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
            self.request.headers.get("X-Forwarded-For") or \
            self.request.remote_ip

    def check_origin(self, origin):
        return True

    def _request_summary(self) -> str:
        client_ip = self.request.headers.get('X-Real-IP') or\
            self.request.headers.get('X-Forwarded-For', '').split(',')[0].strip() or\
            self.request.remote_ip
        summary = "{0} {1} ({2})".format(
            self.request.method,
            self.request.uri,
            client_ip,
        )
        if hasattr(self, "phanterpwa_current_user") and self.phanterpwa_current_user is not None:
            summary = "{0} {1} ({2} - {3})".format(
                self.request.method,
                self.request.uri,
                client_ip,
                self.phanterpwa_current_user.email
            )
        return summary

    @check_user_token()
    def get(self, *args, **kargs):
        q_client = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
        remember_me = q_client.remember_me
        timeout_to_resign = self.projectConfig['BACKEND'][self.app_name]["timeout_to_resign"]
        time_resing_client = int(
            self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire'] * 0.8) + 1
        if remember_me:
            timeout_to_resign = int(
                self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me'] * 0.8) + 1
        t_client = Serialize(
            self.projectConfig['BACKEND'][self.app_name]['secret_key'],
            self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
        )
        t_url = URLSafeSerializer(
            self.projectConfig['BACKEND'][self.app_name]['secret_key'],
            salt="url_secret_key"
        )
        if q_client.last_resign is None or ((q_client.last_resign +
                    timedelta(seconds=timeout_to_resign)) < datetime.now()):
            timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
            if remember_me:
                timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me']
            t_user = Serialize(
                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                timeout_token_user
            )
            content_client = {
                'id_user': str(self.phanterpwa_current_user.id),
                "id_client": str(q_client.id),
                'user_agent': self.phanterpwa_user_agent,
                'remote_addr': self.phanterpwa_remote_ip,
            }
            content_user = {
                'id': str(self.phanterpwa_current_user.id),
                'email': self.phanterpwa_current_user.email
            }
            token_user = t_user.dumps(content_user)
            token_client = self.phanterpwa_client_token
            token_url = t_url.dumps(content_client)
            msg = 'Re-sign just user token'
            q_client.update_record(
                last_resign=datetime.now()
            )
            if not self.phanterpwa_current_user.permit_mult_login:
                self.DALDatabase(
                    (self.DALDatabase.client.auth_user == self.phanterpwa_current_user.id) &
                    (self.DALDatabase.client.token != self.phanterpwa_client_token)
                ).delete()
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                'message': msg,
                'authorization': token_user,
                'client_token': token_client,
                'url_token': token_url,
                "i18n": {
                    "message": self.T(msg)
                }
            })
        elif q_client.date_created +\
                timedelta(seconds=time_resing_client) < datetime.now():
            q_client.delete_record()
            timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
            if remember_me:
                timeout_token_user = self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire_remember_me']
            t_user = Serialize(
                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                timeout_token_user
            )
            new_client = self.DALDatabase.client.insert(
                auth_user=q_client.auth_user,
                remember_me=q_client.remember_me,
                last_resign=datetime.now(),
                date_created=datetime.now()
            )
            content = {
                'id_user': str(self.phanterpwa_current_user.id),
                "id_client": str(new_client),
                'user_agent': self.phanterpwa_user_agent,
                'remote_addr': self.phanterpwa_remote_ip,
            }
            content_user = {
                'id': str(self.phanterpwa_current_user.id),
                'email': self.phanterpwa_current_user.email
            }
            token_user = t_user.dumps(content_user)
            token_url = t_url.dumps(content)
            token_client = t_client.dumps(content)
            q_client.delete_record()
            q_client = self.DALDatabase(self.DALDatabase.client.id == new_client).select().first()
            q_client.update_record(
                token=token_client
            )
            self.phanterpwa_client_token = token_client
            msg = 'Re-sign client token, user token and url token'
            if not self.phanterpwa_current_user.permit_mult_login:
                self.DALDatabase(
                    (self.DALDatabase.client.auth_user == self.phanterpwa_current_user.id) &
                    (self.DALDatabase.client.token != self.phanterpwa_client_token)
                ).delete()
            self.DALDatabase.commit()
            self.set_status(200)
            return self.write({
                "status": "OK",
                'message': msg,
                'authorization': token_user,
                'client_token': token_client,
                'url_token': token_url,
                "i18n": {
                    "message": self.T(msg)
                }
            })
        else:
            msg = "This is not the time to re-sign"
            self.set_status(202)
            return self.write({
                'status': 'Accepted',
                'message': msg,
                "i18n": {
                    "message": self.T(msg)
                }
            })

    def options(self, *args):
        self.set_status(200)
        self.write({"status": "OK"})



import phanterpwa.frontend.helpers as helpers
import phanterpwa.frontend.preloaders as preloaders
from org.transcrypt.stubs.browser import __pragma__

__pragma__('alias', "jQuery", "$")
__pragma__('skip')
jQuery = sessionStorage = JSON = js_undefined = setTimeout = window = this = console =\
    localStorage = 0
__pragma__('noskip')

DIV = helpers.XmlConstructor.tagger("div")
SPAN = helpers.XmlConstructor.tagger("span")
I = helpers.XmlConstructor.tagger("i")
H2 = helpers.XmlConstructor.tagger("h2")
LABEL = helpers.XmlConstructor.tagger("label", False)

__pragma__('kwargs')


class Centralizer(helpers.XmlConstructor):
    def __init__(self, identifier, *content, **attributes):
        attributes["_id"] = "phanterpwa-snippet-{0}".format(identifier)
        if "_class" in attributes:
            attributes["_class"] = "{0}{1}".format(attributes["_class"], " phanterpwa-snippet-centralizer")
        else:
            attributes["_class"] = "phanterpwa-snippet-centralizer"
        self._default_height = attributes.get("default_height", 20)
        if "_style" not in attributes:
            attributes["_style"] = "height:{0}px; min-height: {0}px;".format(self._default_height)

        self._centralizer_html = DIV(
            DIV(
                DIV(
                    DIV(
                        *content,
                        _class="phanterpwa-centralizer-center"
                    ),
                    _class="phanterpwa-centralizer-horizontal"

                ),
                _class="phanterpwa-centralizer-vertical"
            ),
            _class="phanterpwa-centralizer-wrapper"
        )

        helpers.XmlConstructor.__init__(self, 'div', False, self._centralizer_html, **attributes)

    def append(self, value):
        self._centralizer_html.append(value)

    def insert(self, pos, value):
        self._centralizer_html.insert(pos, value)


class PromoOption(helpers.XmlConstructor):
    def __init__(self, identifier, icon, title, description=None, **attributes):
        attributes["_id"] = "phanterpwa-snippet-{0}".format(identifier)
        if "_class" in attributes:
            attributes["_class"] = "{0}{1}".format(attributes["_class"], " phanterpwa-snippet-promooption link p-col w1p100")
        else:
            attributes["_class"] = "phanterpwa-snippet-promooption link p-col w1p100"
        html = DIV(
            DIV(
                icon,
                H2(title, _class="promo-title"),
                _class="promo-icon_and_title"
            ),
            **{"_class": "promo-container"}
        )
        if isinstance(description, helpers.XmlConstructor):
            html.append(DIV(description, _class='promo-content'))
        elif not (description is None or description == ""):
            html.append(DIV(description, _class='promo-content'))
        tag = "div"
        if "_href" in attributes:
            tag = "a"
        helpers.XmlConstructor.__init__(self, tag, False, html, **attributes)


class ICombine(helpers.XmlConstructor):
    def __init__(self, identifier, icon1, icon2,  **attributes):
        attributes["_id"] = "phanterpwa-snippet-{0}".format(identifier)
        if "_class" in attributes:
            attributes["_class"] = "{0}{1}".format(attributes["_class"], " phanterpwa-snippet-icombine")
        else:
            attributes["_class"] = "phanterpwa-snippet-icombine"
        if (isinstance(icon1, I) and isinstance(icon2, I)):
            html = DIV(
                SPAN(icon1, _class="icombine-container-first"),
                SPAN(icon2, _class="icombine-container-last"),
                **{"_class": "icombine-container"}
            )
        else:
            html = DIV(
                "",
                **{"_class": "icombine-container"}
            )
            if window.PhanterPWA.DEBUG:
                console.erro("The icon1 and icon2 must be I instance")
        tag = "div"
        if "_href" in attributes:
            tag = "a"
        helpers.XmlConstructor.__init__(self, tag, False, html, **attributes)


class TogglePanel(helpers.XmlConstructor):
    def __init__(self, *args, **attributes):
        _id = attributes.get('_id', None)
        if _id is None:
            self._namespace = "toggle_panel_{}".format(window.PhanterPWA.get_id())
        else:
            self._namespace = "{}_{}".format(str(_id), window.PhanterPWA.get_id())
        _class = attributes.get("_class")
        _class_wrapper = "phanterpwa-snippet-toggle_panel-wrapper"
        if _class is None:
            attributes['_class'] = "phanterpwa-snippet-toggle_panel"
        else:
            attributes['_class'] = "{0} phanterpwa-snippet-toggle_panel".format(_class)
        toggle_show = attributes.get("toggle_show", True)
        toggle_text_show = attributes.get("toggle_text_show")
        if toggle_text_show is None:
            toggle_text_show = "SHOW"
        toggle_text_hidden = attributes.get("toggle_text_hidden")
        if toggle_text_hidden is None:
            toggle_text_hidden = "HIDDEN"
        toggle_text = toggle_text_hidden
        if not toggle_show:
            _class_wrapper = "{0}_{1}".format(_class_wrapper, "snippet-toggle_panel-hidden")
            toggle_text = toggle_text_hidden
        wrapper_html = DIV(
            DIV(*args, _class="phanterpwa-snippet-toggle_panel-content"),
            DIV(
                DIV(_class="phanterpwa-snippet-toggle_panel-separador"),
                DIV(toggle_text, _class="phanterpwa-snippet-toggle_panel-text"),
                DIV(I(_class="fas fa-chevron-circle-up"), _class="phanterpwa-snippet-toggle_panel-icon"),
                _class="phanterpwa-snippet-toggle_panel-toggle_button link",
                **{"_data-target": self._namespace}
            ),
            _id=self._namespace,
            _class=_class_wrapper
        )
        helpers.XmlConstructor.__init__(self, "div", False, wrapper_html, **attributes)


class Panel(helpers.XmlConstructor):
    def __init__(self, *args, **attributes):
        _id = attributes.get('_id', None)
        if _id is None:
            self._namespace = "panel_{}".format(window.PhanterPWA.get_id())
        else:
            self._namespace = "{}_{}".format(str(_id), window.PhanterPWA.get_id())
        _class = attributes.get("_class")
        _class_wrapper = "phanterpwa-snippet-panel-wrapper"
        _loading = attributes.get("loading", True)
        _icons_button = attributes.get("icons_button", [])
        _label = attributes.get("label", "")
        _header = attributes.get("header", "")
        _footer = attributes.get("footer", "")
        html = []
        if _class is None:
            attributes['_class'] = "phanterpwa-snippet-panel"
        else:
            attributes['_class'] = "{0} phanterpwa-snippet-panel".format(_class)
        if _loading:
            _class_wrapper = "{0} {1}".format(_class_wrapper, "snippets-panel-loading")
        has_icons_button = False
        if isinstance(_icons_button, list):
            if len(_icons_button) > 0:
                has_icons_button = True
                _class_wrapper = "{0} {1}".format(_class_wrapper, "snippets-panel-has_icon_buttons")
        if not(_label is js_undefined or _label is None or _label is ""):
            _label = DIV(
                DIV(LABEL(_label, _for="#{}".format(self._namespace)), _class="phanterpwa-snippet-panel-label-content"),
                DIV(preloaders.indefined_text, _class="phanterpwa-snippet-panel-label-preloader"),
                _class="phanterpwa-snippet-panel-label"
            )
            if has_icons_button:
                _label.append(
                    DIV(
                        I(_class="fas fa-bars"),
                        _id="{}_{}".format(self._namespace, "menu_button"),
                        _class="icon_button phanterpwa-snippet-panel-menu_button",
                        **{"_data-target": self._namespace}
                    )
                )
            _class_wrapper = "{0} {1}".format(_class_wrapper, "snippets-panel-has_label")
            html.append(_label)
        elif has_icons_button:
            _label = DIV(
                DIV(
                    I(_class="fas fa-bars"),
                    _id="{}_{}".format(self._namespace, "menu_button"),
                    _class="icon_button phanterpwa-snippet-panel-menu_button",
                    **{"_data-target": self._namespace}
                ),
                _class="phanterpwa-snippet-panel-label"
            )
        html_content = DIV(_class="phanterpwa-snippet-panel-content")
        if not(_header is js_undefined or _header is None or _header is ""):
            html_content.append(
                DIV(
                    DIV(_header, _class="phanterpwa-snippet-panel-container-header-content"),
                    DIV(preloaders.discs, _class="phanterpwa-snippet-panel-container-header-preloader"),

                    _class="phanterpwa-snippet-panel-container-header"
                )
            )
        html_content.append(DIV(*args, _class="phanterpwa-snippet-panel-container-content"))
        html_content.append(DIV(preloaders.android, _class="phanterpwa-snippet-panel-container-preloader"))

        if has_icons_button:
            xml_icons = DIV(_class="phanterpwa-snippet-panel-icon_buttons-container")
            for x in _icons_button:
                xml_icons.append(DIV(DIV(x, _class="phanterpwa-snippet-panel-icon_button"), DIV(preloaders.indefined_text, _class="phanterpwa-snippet-panel-icon_buttons-preloader", _style="width:40px;")))
            html_content.append(xml_icons)
        html.append(html_content)
        if not(_footer is js_undefined or _footer is None or _footer is ""):
            html.append(
                DIV(
                    DIV(_footer, _class="phanterpwa-snippet-panel-container-footer-content"),
                    DIV(preloaders.discs, _class="phanterpwa-snippet-panel-container-footer-preloader"),
                    _class="phanterpwa-snippet-panel-container-footer"
                )
            )
        wrapper_html = DIV(
            *html,
            _id=self._namespace,
            _class=_class_wrapper
        )
        helpers.XmlConstructor.__init__(self, "div", False, wrapper_html, **attributes)


__pragma__('nokwargs')
